# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                           documentation
#        :author: Charlie Vanaret, Benoit Pauwels, Francois Gallard
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""Design space.

A design space is used to represent the optimization's unknowns,
a.k.a. design variables.

A :class:`.DesignSpace` describes this design space at a given state,
in terms of names, sizes, types, bounds and current values of the design variables.

Variables can easily be added to the :class:`.DesignSpace`
using the :meth:`.DesignSpace.add_variable` method
or removed using the :meth:`.DesignSpace.remove_variable` method.

We can also filter the design variables using the :meth:`.DesignSpace.filter` method.

Getters and setters are also available
to get or set the value of a given variable property.

Lastly,
an instance of :class:`.DesignSpace` can be stored in a txt or HDF file.
"""

from __future__ import division, unicode_literals

import collections
import logging
import sys
from copy import deepcopy
from typing import Any, Dict, Iterable, List, Mapping, Optional, Sequence, Tuple, Union

import h5py
from numpy import abs as np_abs
from numpy import (
    array,
    atleast_1d,
    complex128,
    concatenate,
    empty,
    equal,
    finfo,
    float64,
    genfromtxt,
    inf,
    int32,
    isnan,
    logical_or,
    mod,
    ndarray,
    ones,
    ones_like,
)
from numpy import round_ as np_round
from numpy import string_, vectorize, where, zeros_like
from six import string_types

from gemseo.algos.opt_result import OptimizationResult
from gemseo.core.cache import hash_data_dict
from gemseo.third_party.prettytable import PrettyTable
from gemseo.utils.base_enum import BaseEnum
from gemseo.utils.data_conversion import flatten_mapping
from gemseo.utils.hdf5 import get_hdf5_group
from gemseo.utils.py23_compat import Path, string_array, strings_to_unicode_list

LOGGER = logging.getLogger(__name__)


class DesignVariableType(BaseEnum):
    """A type of design variable."""

    FLOAT = "float"
    INTEGER = "integer"


VarType = Union[
    str,
    Sequence[str],
    DesignVariableType,
    Sequence[DesignVariableType],
]

if sys.version_info < (3, 7, 0):
    DesignVariable = collections.namedtuple(
        "DesignVariable", ["size", "var_type", "l_b", "u_b", "value"]
    )
    DesignVariable.__new__.__defaults__ = (
        1,
        DesignVariableType.FLOAT,
        None,
        None,
        None,
    )
else:
    DesignVariable = collections.namedtuple(
        "DesignVariable",
        ["size", "var_type", "l_b", "u_b", "value"],
        defaults=(
            1,
            DesignVariableType.FLOAT,
            None,
            None,
            None,
        ),
    )


class DesignSpace(collections.MutableMapping):
    """Description of a design space.

    It defines a set of variables from their names, sizes, types and bounds.

    In addition,
    it provides the current values of these variables
    that can be used as the initial solution of an :class:`.OptimizationProblem`.

    A :class:`.DesignSpace` has the same API as a dictionary,
    e.g. ``variable = design_space["x"]``,
    ``other_design_space["x"] = design_space["x"]``,
    ``del design_space["x"]``,
    ``for name, value in design_space["x"].items()``, ...

    Attributes:
        name (Optional[str]): The name of the space.
        variables_names (List[str]): The names of the variables.
        dimension (int): The total dimension of the space,
            corresponding to the sum of the sizes of the variables.
        variables_sizes (Dict[str,int]): The sizes of the variables.
        variables_types (Dict[str,ndarray]): The types of the variables components,
            which can be any :attr:`DesignVariableType`.
        normalize (Dict[str,ndarray]): The normalization policies
            of the variables components indexed by the variables names;
            if `True`, the component can be normalized.
    """

    FLOAT = DesignVariableType.FLOAT
    INTEGER = DesignVariableType.INTEGER
    AVAILABLE_TYPES = [FLOAT, INTEGER]
    __TYPE_NAMES = tuple(x.value for x in DesignVariableType)
    __TYPES_TO_DTYPES = {
        DesignVariableType.FLOAT.value: "float64",
        DesignVariableType.INTEGER.value: "int32",
    }
    MINIMAL_FIELDS = ["name", "lower_bound", "upper_bound"]
    TABLE_NAMES = ["name", "lower_bound", "value", "upper_bound", "type"]

    DESIGN_SPACE_GROUP = "design_space"
    NAME_GROUP = "name"
    NAMES_GROUP = "names"
    LB_GROUP = "l_b"
    UB_GROUP = "u_b"
    VAR_TYPE_GROUP = "var_type"
    VALUE_GROUP = "value"
    SIZE_GROUP = "size"
    # separator that denotes a vector's components
    SEP = "!"

    def __init__(
        self,
        hdf_file=None,  # type: Optional[Union[str,Path]]
        name=None,  # type: Optional[str]
    ):  # type: (...) -> None
        """
        Args:
            hdf_file: The path to the file
                containing the description of an initial design space.
                If None, start with an empty design space.
            name: The name to be given to the design space,
                `None` if the design space is unnamed.
        """
        self.name = name
        self.variables_names = []
        self.dimension = 0
        self.variables_sizes = {}
        self.variables_types = {}
        self.normalize = {}
        self._lower_bounds = {}
        self._upper_bounds = {}
        # These attributes are stored for faster computation of normalization
        # and unnormalization
        self._norm_factor = None
        self.__lower_bounds_array = None
        self.__upper_bounds_array = None
        self.__int_vars_indices = None
        self.__norm_data_is_computed = False
        self.__norm_inds = None
        self.__to_zero = None
        self.__bound_tol = 100.0 * finfo(float64).eps
        self._current_x = {}
        if hdf_file is not None:
            self.import_hdf(hdf_file)

    def __delitem__(
        self,
        name,  # type: str
    ):  # type: (...) -> None
        """Remove a variable from the design space.

        Args:
            name: The name of the variable to be removed.
        """
        self.remove_variable(name)

    def remove_variable(
        self,
        name,  # type: str
    ):  # type: (...) -> None
        """Remove a variable from the design space.

        Args:
            name: The name of the variable to be removed.
        """
        self.__norm_data_is_computed = False
        size = self.variables_sizes[name]
        self.dimension -= size
        self.variables_names.remove(name)
        del self.variables_sizes[name]
        del self.variables_types[name]
        del self.normalize[name]
        if name in self._lower_bounds:
            del self._lower_bounds[name]
        if name in self._upper_bounds:
            del self._upper_bounds[name]
        if name in self._current_x:
            del self._current_x[name]

    def filter(
        self,
        keep_variables,  # type: Union[str,Iterable[str]]
        copy=False,  # type: bool
    ):  # type: (...) -> DesignSpace
        """Filter the design space to keep a subset of variables.

        Args:
            keep_variables: The names of the variables to be kept.
            copy: If True, then a copy of the design space is filtered,
                otherwise the design space itself is filtered.

        Returns:
            Either the filtered original design space or a copy.

        Raises:
            ValueError: If the variable is not in the design space.
        """
        if isinstance(keep_variables, string_types):
            keep_variables = [keep_variables]
        design_space = deepcopy(self) if copy else self
        for name in deepcopy(self.variables_names):
            if name not in keep_variables:
                design_space.remove_variable(name)
        for name in keep_variables:
            if name not in self.variables_names:
                raise ValueError("Variable '{}' is not known.".format(name))
        return design_space

    def filter_dim(
        self,
        variable,  # type: str
        keep_dimensions,  # type: Iterable[int]
    ):  # type: (...) -> DesignSpace
        """Filter the design space to keep a subset of dimensions for a variable.

        Args:
            variable: The name of the variable.
            keep_dimensions: The dimensions of the variable to be kept,
                between :math:`0` and :math:`d-1`
                where :math:`d` is the number of dimensions of the variable.

        Returns:
            The filtered design space.

        Raises:
            ValueError: If a dimension is unknown.
        """
        self.__norm_data_is_computed = False
        removed_dimensions = list(
            set(range(self.variables_sizes[variable])) - set(keep_dimensions)
        )
        bad_dimensions = list(
            set(keep_dimensions) - set(range(self.variables_sizes[variable]))
        )
        size = len(removed_dimensions)
        self.dimension -= size
        self.variables_sizes[variable] -= size
        types = []
        for dimension in keep_dimensions:
            if dimension in bad_dimensions:
                self.remove_variable(variable)
                raise ValueError(
                    "Dimension {} of variable '{}' is not known.".format(
                        dimension, variable
                    )
                )
            types.append(self.variables_types[variable][dimension])
            self.variables_types[variable] = array(types)
        idx = keep_dimensions
        self.normalize[variable] = self.normalize[variable][idx]
        if variable in self._lower_bounds:
            self._lower_bounds[variable] = self._lower_bounds[variable][idx]
        if variable in self._upper_bounds:
            self._upper_bounds[variable] = self._upper_bounds[variable][idx]
        if variable in self._current_x:
            self._current_x[variable] = self._current_x[variable][idx]
        return self

    def add_variable(
        self,
        name,  # type: str
        size=1,  # type: int
        var_type=DesignVariableType.FLOAT,  # type: VarType
        l_b=None,  # type: Optional[Union[float,ndarray]]
        u_b=None,  # type: Optional[Union[float,ndarray]]
        value=None,  # type:  Optional[Union[float,ndarray]]
    ):  # type: (...) -> None
        r"""Add a variable to the design space.

        Args:
            name: The name of the variable.
            size: The size of the variable.
            var_type: Either the type of the variable
                or the types of its components.
            l_b: The lower bound of the variable.
                If None, use :math:`-\infty`.
            u_b: The upper bound of the variable.
                If None, use :math:`+\infty`.
            value: The default value of the variable.
                If None, do not use a default value.

        Raises:
            ValueError: Either if the variable already exists
                or if the size is not a positive integer.
        """
        if name in self.variables_names:
            raise ValueError("Variable '{}' already exists.".format(name))

        if size <= 0 or int(size) != size:
            raise ValueError(
                "The size of '{}' should be a positive integer.".format(name)
            )

        # name and size
        self.variables_names.append(name)
        self.dimension += size
        self.variables_sizes[name] = size
        # type
        self._add_type(name, size, var_type)

        # bounds
        self._add_bound(name, size, l_b, is_lower=True)
        self._add_bound(name, size, u_b, is_lower=False)
        self._check_variable_bounds(name)

        # normalization policy
        self._add_norm_policy(name)

        if value is not None:
            array_value = atleast_1d(value)
            self._check_value(array_value, name)
            if len(array_value) == 1 and size > 1:
                array_value = array_value * ones(size)
            self._current_x[name] = array_value
            self._check_current_x_value(name)

    def _add_type(
        self,
        name,  # type: str
        size,  # type: int
        var_type=DesignVariableType.FLOAT,  # type: VarType
    ):  # type: (...) -> None
        """Add a type to a variable.

        Args:
            name: The name of the variable.
            size: The size of the variable.
            var_type: Either the type of the variable (see :attr:`AVAILABLE_TYPES`)
                or the types of its components.

        Raises:
            ValueError: Either if the number of component types is different
                from the variable size or if a variable type is unknown.
        """
        if not hasattr(var_type, "__iter__") or isinstance(
            var_type, (string_types, DesignVariableType, bytes)
        ):
            var_type = [var_type] * size

        if len(var_type) != size:
            raise ValueError(
                "The list of types for variable '{}' should be of size {}.".format(
                    name, size
                )
            )

        var_types = []

        for v_type in var_type:
            if isinstance(v_type, bytes):
                v_type = v_type.decode()

            if isinstance(v_type, string_types) and v_type not in self.__TYPE_NAMES:
                msg = 'The type "{0}" of {1} is not known.'.format(v_type, name)
                raise ValueError(msg)
            elif v_type in DesignVariableType:
                v_type = v_type.value

            var_types += [v_type]

        self.variables_types[name] = array(var_types)
        self.__norm_data_is_computed = False

    def _add_norm_policy(
        self,
        name,  # type: str
    ):  # type: (...) -> None
        """Add a normalization policy to a variable.

        Unbounded variables are not normalized.
        Bounded variables (both from above and from below) are normalized.

        Args:
            name: The name of a variable.

        Raises:
            ValueError: Either if the variable is not in the design space,
                it its size is not set,
                if the types of its components are not set
                or if there is no implemented normalization policy
                for the type of this variable.
        """
        # Check that the variable is in the design space:
        if name not in self.variables_names:
            raise ValueError("Variable '{}' is not known.".format(name))
        # Check that the variable size is set:
        size = self.get_size(name)
        if size is None:
            raise ValueError("The size of variable '{}' is not set.".format(name))
        # Check that the variables types are set:
        variables_types = self.variables_types.get(name, None)
        if variables_types is None:
            raise ValueError(
                "The components types of variable '{}' are not set.".format(name)
            )
        # Set the normalization policy:
        normalize = empty(size)
        for i in range(size):
            var_type = variables_types[i]
            if var_type in self.__TYPES_TO_DTYPES:
                if (
                    self._lower_bounds[name][i] == -inf
                    or self._upper_bounds[name][i] == inf
                ):
                    # Unbounded variables are not normalized:
                    normalize[i] = False
                elif self._lower_bounds[name][i] == self._upper_bounds[name][i]:
                    # Constant variables are not normalized:
                    normalize[i] = False
                else:
                    normalize[i] = True
            else:
                msg = "The normalization policy for type {0} is not implemented."
                raise ValueError(msg.format(var_type))
        self.normalize[name] = normalize

    @staticmethod
    def __is_integer(
        value,  # type: ndarray
    ):  # type: (...) -> bool
        """Check that all values in an array are integers.

        Args:
            value: The array to be checked.

        Returns:
            Whether all values in the array are integers.
        """
        are_none = equal(value, None)
        are_int = equal(mod(value.astype("f"), 1), 0)
        return logical_or(are_none, are_int)

    @staticmethod
    def __is_numeric(
        value,  # type: Any
    ):  # type: (...) -> bool
        """Check that a value is numeric.

        Args:
            value: The value to be checked.

        Returns:
            Whether the value is numeric.
        """
        res = (value is None) or hasattr(value, "real")
        try:
            if not res:
                float(value)
            return True
        except TypeError:
            return False

    @staticmethod
    def __is_not_nan(
        value,  # type: ndarray
    ):  # type: (...) -> bool
        """Check that a value is not a nan.

        Args:
            value: The value to be checked.

        Returns:
            Whether the value is not a nan.
        """
        return (value is None) or ~isnan(value)

    def _check_value(
        self,
        value,  # type: ndarray
        name,  # type: str
    ):  # type: (...) -> bool
        """Check that the value of a variable is valid.

        Args:
            value: The value to be checked.
            name: The name of the variable.

        Returns:
            Whether the value of the variable is valid.

        Raises:
            ValueError: Either if the array is not one-dimensional,
                if the value is not numerizable,
                if the value is nan
                or if there is a component value which is not an integer
                while the variable type is integer.
        """
        all_indices = set(range(len(value)))
        # OK if the variable value is one-dimensional
        if len(value.shape) > 1:
            raise ValueError(
                "Value {} of variable '{}' has dimension greater than 1 "
                "while a float or a 1d iterable object "
                "(array, list, tuple, ...) "
                "while a scalar was expected.".format(value, name)
            )

        # OK if all components are None
        if all(equal(value, None)):
            return True

        test = vectorize(self.__is_numeric)(value)
        indices = all_indices - set(list(where(test)[0]))
        for idx in indices:
            raise ValueError(
                "Value {} of variable '{}' is not numerizable.".format(value[idx], name)
            )

        test = vectorize(self.__is_not_nan)(value)
        indices = all_indices - set(list(where(test)[0]))
        for idx in indices:
            raise ValueError(
                "Value {} of variable '{}' is nan.".format(value[idx], name)
            )

        # Check if some components of an integer variable are not integer.
        if self.variables_types[name][0] == DesignVariableType.INTEGER.value:
            indices = all_indices - set(list(where(self.__is_integer(value))[0]))
            for idx in indices:
                raise ValueError(
                    "Component value {} of variable '{}' is not an integer "
                    "while variable is of type integer "
                    "(index: {}).".format(value[idx], name, idx)
                )

    def _add_bound(
        self,
        name,  # type: str
        size,  # type: int
        bound,  # type: ndarray
        is_lower=True,  # type: bool
    ):  # type: (...) -> None
        """Add a lower or upper bound to a variable.

        Args:
            name: The name of the variable.
            size: The size of the variable.
            bound: The bound of the variable.
            is_lower: If True, the bound is a lower bound.
                Otherwise, it is an upper bound.

        Raises:
            ValueError: If the size of the bound is different
                from the size of the variable.
        """
        self.__norm_data_is_computed = False
        if is_lower:
            bound_dict = self._lower_bounds
        else:
            bound_dict = self._upper_bounds
        infinity = inf * ones(size)
        if bound is None:
            if is_lower:
                bound_array = -infinity
            else:
                bound_array = infinity
            bound_dict.update({name: bound_array})
            return
        bound_array = atleast_1d(bound)
        self._check_value(bound_array, name)
        if hasattr(bound, "__iter__"):
            # iterable structure
            if len(bound_array) != size:
                bound_str = "lower" if is_lower else "upper"
                raise ValueError(
                    "The {} bounds of '{}' should be of size {}.".format(
                        bound_str, name, size
                    )
                )
            if is_lower:
                bound_array = where(equal(bound_array, None), -infinity, bound_array)
            else:
                bound_array = where(equal(bound_array, None), infinity, bound_array)
            bound_dict.update({name: bound_array})
        else:
            # scalar: same lower bound for all components
            bound_array = bound * ones(size)
            bound_dict.update({name: bound_array})
        return

    def _check_variable_bounds(
        self,
        name,  # type: str
    ):  # type: (...) -> None
        """Check that the bounds of a variable are compatible and have the same size.

        Args:
            name: The name of the variable.

        Raises:
            ValueError: If the bounds of the variable are not valid.
        """
        l_b = self._lower_bounds.get(name, None)
        u_b = self._upper_bounds.get(name, None)
        inds = where(u_b < l_b)[0]
        if inds.size != 0:
            raise ValueError(
                "The bounds of variable '{}'{} are not valid: {}!<{}.".format(
                    name, inds, u_b[inds], l_b[inds]
                )
            )

    def _check_current_x_value(
        self,
        name,  # type: str
    ):  # type: (...) -> None
        """Check that the current value of a variable is between its bounds.

        Args:
            name: The name of the variable.

        Raises:
            ValueError: If the current value of the variable is outside its bounds.
        """
        l_b = self._lower_bounds.get(name, None)
        u_b = self._upper_bounds.get(name, None)
        c_x = self._current_x.get(name, None)
        not_none = ~equal(c_x, None)
        inds = where(
            logical_or(
                c_x[not_none] < l_b[not_none] - 1e-14,
                c_x[not_none] > u_b[not_none] + 1e-14,
            )
        )[0]
        for idx in inds:
            raise ValueError(
                "The current value of variable '{}' ({}) is not "
                "between the lower bound {} and the upper bound {}.".format(
                    name, c_x[idx], l_b[idx], u_b[idx]
                )
            )

    def has_current_x(self):  # type: (...) -> bool
        """Check if the current design value is defined for all variables.

        Returns:
            Whether the current design value is defined for all variables.
        """
        if self._current_x is None or len(self._current_x) != len(self.variables_names):
            return False
        for val in self._current_x.values():
            if val is None:
                return False
        return True

    def check(self):  # type: (...) -> None
        """Check the state of the design space.

        Raises:
            ValueError: If the design space is empty.
        """
        if not self.variables_names:
            raise ValueError("The design space is empty.")

        for name in self.variables_names:
            self._check_variable_bounds(name)

        if self.has_current_x():
            self._check_current_x()

    def check_membership(
        self,
        x_vect,  # type: Union[Mapping[str,ndarray],ndarray]
        variables_names=None,  # type: Optional[Sequence[str]]
    ):  # type: (...) -> None
        """Check whether the variables satisfy the design space requirements.

        Args:
            x_vect: The values of the variables.
            variables_names: The names of the variables.
                If None, use the names of the variables of the design space.

        Raises:
            ValueError: Either if the dimension of the values vector is wrong,
                if the values are not specified as an array or a dictionary,
                if the values are outside the bounds of the variables or
                if the component of an integer variable is an integer.
        """
        # Convert the input vector into a dictionary if necessary:
        if isinstance(x_vect, dict):
            x_dict = x_vect
        elif isinstance(x_vect, ndarray):
            if x_vect.size != self.dimension:
                raise ValueError(
                    "The dimension of the input array ({}) should be {}.".format(
                        x_vect.size, self.dimension
                    )
                )
            x_dict = self.array_to_dict(x_vect)
        else:
            raise TypeError(
                "The input vector should be an array or a dictionary; "
                "got {} instead.".format(type(x_vect))
            )
        # Check the membership of the input vector to the design space:
        variables_names = variables_names or self.variables_names
        for name in variables_names:
            if x_dict[name] is None:
                continue
            size = self.variables_sizes[name]
            l_b = self._lower_bounds.get(name, None)
            u_b = self._upper_bounds.get(name, None)

            if atleast_1d(x_dict[name]).size != size:
                raise ValueError(
                    "The component '{}' of the given array should have size {}.".format(
                        name, size
                    )
                )
            for i in range(size):
                x_real = atleast_1d(x_dict[name])[i].real
                if l_b is not None and x_real < l_b[i] - self.__bound_tol:
                    msg = (
                        "The component {}{}{} of the given array ({}) "
                        "is lower than the lower bound ({}) by {:.1e}.".format(
                            name, self.SEP, i, x_real, l_b[i], l_b[i] - x_real
                        )
                    )
                    raise ValueError(msg)
                if u_b is not None and u_b[i] + self.__bound_tol < x_real:
                    msg = (
                        "The component '{}'{}{} of the given array ({}) "
                        "is greater than the upper bound ({}) by {:.1e}.".format(
                            name, self.SEP, i, x_real, u_b[i], u_b[i] - x_real
                        )
                    )
                    raise ValueError(msg)
                if (
                    self.variables_types[name][0] == DesignVariableType.INTEGER.value
                ) and not self.__is_integer(x_real):
                    msg = (
                        "The component '{}'{}{} of the given array is not an integer "
                        "while variable is of type integer! Value = {}".format(
                            name, self.SEP, i, x_real
                        )
                    )
                    raise ValueError(msg)

    def get_active_bounds(
        self,
        x_vec=None,  # type: Optional[ndarray]
        tol=1e-8,  # type: float
    ):  # type: (...) -> Tuple[Dict[str,ndarray],Dict[str,ndarray]]
        """Determine which bound constraints of the current point are active.

        Args:
            x_vec: The point at which to check the bounds.
                If None, use the current point.
            tol: The tolerance of comparison of a scalar with a bound.

        Returns:
            Whether the components of the lower and upper bound constraints are active,
            the first returned value representing the lower bounds
            and the second one the upper bounds, e.g.

            .. code-block:: python

                   ({'x': array(are_x_lower_bounds_active),
                     'y': array(are_y_lower_bounds_active)},
                    {'x': array(are_x_upper_bounds_active),
                     'y': array(are_y_upper_bounds_active)}
                   )

            where:

            .. code-block:: python

                are_x_lower_bounds_active = [True, False]
                are_x_upper_bounds_active = [False, False]
                are_y_lower_bounds_active = [False]
                are_y_upper_bounds_active = [True]
        """
        if x_vec is None:
            x_dict = self._current_x
            self.check_membership(self.get_current_x())
        elif isinstance(x_vec, ndarray):
            x_dict = self.array_to_dict(x_vec)
        elif isinstance(x_vec, dict):
            x_dict = x_vec
        else:
            raise TypeError(
                "Expected dict or array for x_vec argument; "
                "got {}.".format(type(x_vec))
            )

        active_l_b = {}
        active_u_b = {}
        for name in self.variables_names:
            l_b = self._lower_bounds.get(name)
            l_b = where(equal(l_b, None), -inf, l_b)
            u_b = self._upper_bounds.get(name)
            u_b = where(equal(u_b, None), inf, u_b)
            x_vec_i = x_dict[name]
            # lower bound saturated
            albi = where(np_abs(x_vec_i - l_b) <= tol, True, False)
            active_l_b[name] = albi
            # upper bound saturated
            aubi = where(np_abs(x_vec_i - u_b) <= tol, True, False)
            active_u_b[name] = aubi
        return active_l_b, active_u_b

    def _check_current_x(
        self,
        variables_names=None,  # type: Optional[Iterable[str]]
    ):  # type: (...) -> None
        """Check the names of the current point.

        Args:
            variables_names: The names of the variables.
                If None, use the names of the variables of the design space.

        Raises:
            ValueError: If the names of the variables of the current point
                and the names of the variables of the design space are different.
        """
        if sorted(set(self.variables_names)) != sorted(set(self._current_x.keys())):
            raise ValueError(
                "Expected current_x variables: {}; got {}.".format(
                    self.variables_names, list(self._current_x.keys())
                )
            )
        self.check_membership(self._current_x, variables_names)

    def get_current_x(
        self,
        variables_names=None,  # type: Optional[Iterable[str]]
    ):  # type: (...) -> ndarray
        """Return the current point in the design space.

        Args:
            variables_names: The names of the required variables.
                If None, use the names of the variables of the design space.

        Raises:
            KeyError: If a variable has no current value.
        """
        try:
            x_arr = self.dict_to_array(self._current_x, all_var_list=variables_names)
            return x_arr
        except KeyError as err:
            raise KeyError(
                "The design space has no current value for '{}'.".format(err.args[0])
            )

    def get_indexed_var_name(
        self,
        variable_name,  # type: str
    ):  # type: (...) -> Union[str,List[str]]
        """Create the names of the components of a variable.

        If the size of the variable is equal to 1,
        this method returns the name of the variable.
        Otherwise,
        it concatenates the name of the variable,
        the separator :attr:`SEP` and the index of the component.

        Args:
            variable_name: The name of the variable.

        Returns:
            The names of the components of the variable.
        """
        size = self.variables_sizes[variable_name]
        if size == 1:
            return variable_name
        return [variable_name + self.SEP + str(i) for i in range(size)]

    def get_indexed_variables_names(self):  # type: (...) -> List[str]
        """Create the names of the components of all the variables.

        If the size of the variable is equal to 1,
        this method uses its name.
        Otherwise,
        it concatenates the name of the variable,
        the separator :attr:`SEP` and the index of the component.

        Returns:
            The name of the components of all the variables.
        """
        var_ind_names = []
        for var in self.variables_names:
            vnames = self.get_indexed_var_name(var)
            if isinstance(vnames, string_types):
                var_ind_names.append(vnames)
            else:
                var_ind_names += vnames
        return var_ind_names

    def get_variables_indexes(
        self,
        variables_names,  # type: Iterable[str]
    ):  # type: (...) -> ndarray
        """Return the indexes of a design array corresponding to the variables names.

        Args:
            variables_names: The names of the variables.

        Returns:
            The indexes of a design array corresponding to the variables names.
        """
        indexes = list()
        index = 0
        for name in self.variables_names:
            var_size = self.get_size(name)
            if name in variables_names:
                indexes.extend(range(index, index + var_size))
            index += var_size
        return array(indexes)

    def __update_normalization_vars(self):  # type: (...) -> None
        """Compute the inner attributes used for normalization and unnormalization."""

        self.__lower_bounds_array = self.get_lower_bounds()
        self.__upper_bounds_array = self.get_upper_bounds()
        self._norm_factor = self.__upper_bounds_array - self.__lower_bounds_array
        norm_array = self.dict_to_array(self.normalize)
        self.__norm_inds = where(norm_array)[0]
        # In case lb=ub
        self.__to_zero = where(self._norm_factor == 0.0)[0]

        var_ind_list = []
        for var in self.variables_names:
            # Store the mask of int variables
            to_one = self.variables_types[var] == DesignVariableType.INTEGER.value
            var_ind_list.append(to_one)
        self.__int_vars_indices = concatenate(var_ind_list)
        self.__norm_data_is_computed = True

    def normalize_vect(
        self,
        x_vect,  # type: ndarray
        minus_lb=True,  # type: bool
    ):  # type: (...) -> ndarray
        r"""Normalize a vector of the design space.

        If `minus_lb` is True:

        .. math::

           x_u = \frac{x-l_b}{u_b-l_b}

        where :math:`l_b` and :math:`u_b` are the lower and upper bounds of :math:`x`.

        Otherwise:

        .. math::

           x_u = \frac{x}{u_b-l_b}

        Unbounded variables are not normalized.

        Args:
            x_vect: The values of the design variables.
            minus_lb: If True, remove the lower bounds at normalization.

        Returns:
            The normalized vector.

        Raises:
            ValueError: If the array to be normalized is not one- or two-dimensional.
        """
        if not self.__norm_data_is_computed:
            self.__update_normalization_vars()

        # Normalize the relevant components:
        if self.has_current_x():
            current_dtype = self.get_current_x().dtype
        else:
            current_dtype = float
        norm_vect = x_vect.astype(current_dtype, copy=True)

        # In case lb=ub
        norm_inds = self.__norm_inds

        if len(x_vect.shape) == 1:
            if minus_lb:
                norm_vect[norm_inds] -= self.__lower_bounds_array[norm_inds]
            norm_vect[norm_inds] /= self._norm_factor[norm_inds]
            # In case lb=ub put value to 0
        elif len(x_vect.shape) == 2:
            if minus_lb:
                norm_vect[:, norm_inds] -= self.__lower_bounds_array[norm_inds]
            norm_vect[:, norm_inds] /= self._norm_factor[norm_inds]
        else:
            raise ValueError("The array to be normalized must be 1d or 2d.")

        to_zero = self.__to_zero
        if to_zero.size > 0:
            norm_vect[to_zero] = 0.0
        return norm_vect

    def normalize_grad(
        self,
        g_vect,  # type: ndarray
    ):  # type: (...) -> ndarray
        r"""Normalize an unnormalized gradient.

        This method is based on the chain rule:

        .. math::

           \frac{df(x)}{dx}
           = \frac{df(x)}{dx_u}\frac{dx_u}{dx}
           = \frac{df(x)}{dx_u}\frac{1}{u_b-l_b}

        where
        :math:`x_u = \frac{x-l_b}{u_b-l_b}` is the normalized input vector,
        :math:`x` is the unnormalized input vector
        and :math:`l_b` and :math:`u_b` are the lower and upper bounds of :math:`x`.

        Then,
        the normalized gradient reads:

        .. math::

           \frac{df(x)}{dx_u} = (u_b-l_b)\frac{df(x)}{dx}

        where :math:`\frac{df(x)}{dx}` is the unnormalized one.

        Args:
            g_vect: The gradient to be normalized.

        Returns:
            The normalized gradient.
        """
        return self.unnormalize_vect(g_vect, minus_lb=False, no_check=True)

    def unnormalize_grad(
        self,
        g_vect,  # type: ndarray
    ):  # type: (...) -> ndarray
        r"""Unnormalize a normalized gradient.

        This method is based on the chain rule:

        .. math::

           \frac{df(x)}{dx}
           = \frac{df(x)}{dx_u}\frac{dx_u}{dx}
           = \frac{df(x)}{dx_u}\frac{1}{u_b-l_b}

        where
        :math:`x_u = \frac{x-l_b}{u_b-l_b}` is the normalized input vector,
        :math:`x` is the unnormalized input vector,
        :math:`\frac{df(x)}{dx_u}` is the unnormalized gradient
        :math:`\frac{df(x)}{dx}` is the normalized one,
        and :math:`l_b` and :math:`u_b` are the lower and upper bounds of :math:`x`.

        Args:
            g_vect: The gradient to be unnormalized.

        Returns:
            The unnormalized gradient.
        """
        return self.normalize_vect(g_vect, minus_lb=False)

    def unnormalize_vect(
        self,
        x_vect,  # type: ndarray
        minus_lb=True,  # type: bool
        no_check=False,  # type: bool
    ):  # type: (...) -> ndarray
        """Unnormalize a normalized vector of the design space.

        If `minus_lb` is True:

        .. math::

           x = x_u(u_b-l_b) + l_b

        where
        :math:`x_u` is the normalized input vector,
        :math:`x` is the unnormalized input vector
        and :math:`l_b` and :math:`u_b` are the lower and upper bounds of :math:`x`.

        Otherwise:

        .. math::

           x = x_u(u_b-l_b)

        Args:
            x_vect: The values of the design variables.
            minus_lb: If True, remove the lower bounds at normalization.
            no_check: If True, do not check that the values are in [0,1].

        Returns:
            The unnormalized vector.

        Raises:
            ValueError: If the array to be unnormalized is not one- or two-dimensional.
        """
        n_dims = x_vect.ndim
        if n_dims not in [1, 2]:
            raise ValueError(
                "The array to be unnormalized must be 1d or 2d, got {}d.".format(n_dims)
            )
        if not self.__norm_data_is_computed:
            self.__update_normalization_vars()
        norm_inds = self.__norm_inds
        l_bounds = self.__lower_bounds_array
        if not no_check:
            # Check whether the input components are between 0 and 1:
            if n_dims == 1:
                bounded = x_vect[norm_inds]
            else:
                bounded = x_vect[:, norm_inds]
            if (bounded < -1e-12).any() or (bounded > 1 + 1e-12).any():
                msg = (
                    "All components of the normalized vector should be between 0 and 1."
                )
                lb_viol = bounded[bounded < -1e-12]
                if lb_viol.size != 0:
                    msg += " Lower bounds violated: {}.".format(lb_viol)
                ub_viol = bounded[bounded > 1 + 1e-12]
                if ub_viol.size != 0:
                    msg += " Upper bounds violated: {}.".format(ub_viol)

                LOGGER.warning(msg)
        # Unnormalize the relevant components:
        if self.has_current_x():
            current_dtype = self.get_current_x().dtype
        else:
            current_dtype = float
        unnorm_vect = x_vect.astype(current_dtype, copy=True)

        if n_dims == 1:
            unnorm_vect[norm_inds] *= self._norm_factor[norm_inds]
            if minus_lb:
                unnorm_vect[norm_inds] += l_bounds[norm_inds]
        else:
            unnorm_vect[:, norm_inds] *= self._norm_factor[norm_inds]
            if minus_lb:
                unnorm_vect[:, norm_inds] += l_bounds[norm_inds]
        inds_fixed = self.__to_zero
        if inds_fixed.size > 0:
            if n_dims == 1:
                unnorm_vect[inds_fixed] = l_bounds[inds_fixed]
            else:
                unnorm_vect[:, inds_fixed] = l_bounds[inds_fixed]

        r_xvec = self.round_vect(unnorm_vect)
        return r_xvec

    def transform_vect(
        self, vector  # type: ndarray
    ):  # type:(...) -> ndarray
        """Map a point of the design space to a vector with components in :math:`[0,1]`.

        Args:
            vector: A point of the design space.

        Returns:
            A vector with components in :math:`[0,1]`.
        """
        return self.normalize_vect(vector)

    def untransform_vect(
        self, vector  # type: ndarray
    ):  # type:(...) -> ndarray
        """Map a vector with components in :math:`[0,1]` to the design space.

        Args:
            vector: A vector with components in :math:`[0,1]`.

        Returns:
            A point of the variables space.
        """
        return self.unnormalize_vect(vector)

    def round_vect(
        self,
        x_vect,  # type: ndarray
    ):  # type: (...) -> ndarray
        """Round the vector where variables are of integer type.

        Args:
            x_vect: The values to be rounded.

        Returns:
            The rounded values values.

        Raises:
            ValueError: If the values is not a one- or two-dimensional
        """
        if not self.__norm_data_is_computed:
            self.__update_normalization_vars()
        int_vars = self.__int_vars_indices
        if not int_vars.any():
            return x_vect
        n_dims = len(x_vect.shape)
        rounded_x_vect = x_vect.copy()
        if n_dims == 1:
            rounded_x_vect[int_vars] = np_round(rounded_x_vect[int_vars])
        elif n_dims == 2:
            rounded_x_vect[:, int_vars] = np_round(rounded_x_vect[:, int_vars])
        else:
            raise ValueError(
                "The array to be unnormalized must be 1d or 2d; got {}d.".format(n_dims)
            )

        return rounded_x_vect

    def get_current_x_normalized(self):  # type: (...) -> ndarray
        """Return the current point normalized.

        Returns:
            The current point as an array normalized by the bounds of the variables.

        Returns:
            KeyError: If the current point cannot be normalized.
        """
        try:
            current_x = self.get_current_x()
        except KeyError as err:
            raise KeyError(
                "Cannot compute normalized current value since {}.".format(err.args[0])
            )
        return self.normalize_vect(current_x)

    def get_current_x_dict(self):  # type:(...) -> Dict[str,ndarray]
        """Return the current point in the design space as a dictionary.

        Returns:
            The current point in the design space as a dictionary,
            whose keys are the names of the variables
            and values are the values of the variables.
        """
        return self._current_x

    def set_current_x(
        self,
        current_x,  # type: Union[ndarray,Mapping[str,ndarray], OptimizationResult]
    ):  # type: (...) -> None
        """Set the current point.

        Args:
            current_x: The value of the current point.

        Raises:
            ValueError: If the value has a wrong dimension.
            TypeError: If the current point is neither a mapping of NumPy arrays,
                a NumPy array nor an :class:`.OptimizationResult`.
        """
        if isinstance(current_x, dict):
            self._current_x = current_x
        elif isinstance(current_x, ndarray):
            if current_x.size != self.dimension:
                raise ValueError(
                    "Invalid current_x, "
                    "dimension mismatch: {} != {}.".format(
                        self.dimension, current_x.size
                    )
                )
            self._current_x = self.array_to_dict(current_x)
        elif isinstance(current_x, OptimizationResult):
            if current_x.x_opt.size != self.dimension:
                raise ValueError(
                    "Invalid x_opt, "
                    "dimension mismatch: {} != {}.".format(
                        self.dimension, current_x.x_opt.size
                    )
                )
            self._current_x = self.array_to_dict(current_x.x_opt)
        else:
            raise TypeError(
                "The current point should be either an array, "
                "a dictionary of arrays "
                "or an optimization result; "
                "got {} instead.".format(type(current_x))
            )

        for name, value in self._current_x.items():
            if value is not None:
                variable_type = self.variables_types[name]
                if isinstance(variable_type, ndarray):
                    variable_type = variable_type[0]
                if variable_type == DesignVariableType.INTEGER.value:
                    value = value.astype(self.__TYPES_TO_DTYPES[variable_type])
                self._current_x[name] = value

        self._check_current_x()

    def set_current_variable(
        self,
        name,  # type: str
        current_value,  # type: ndarray
    ):
        """Set the current value of a single variable.

        Args:
            name: The name of the variable.
            current_value: The current value of the variable.
        """
        if name in self.variables_names:
            self._current_x[name] = current_value
        else:
            raise ValueError("Variable '{}' is not known.".format(name))

    def get_size(
        self,
        name,  # type: str
    ):  # type:(...) -> Optional[int]
        """Get the size of a variable.

        Args:
            name: The name of the variable.

        Returns:
            The size of the variable, None if it is not known.
        """
        return self.variables_sizes.get(name, None)

    def get_type(
        self,
        name,  # type: str
    ):  # type:(...) -> Optional[str]
        """Return the type of a variable.

        Args:
            name: The name of the variable.

        Returns:
            The type of the variable, None if it is not known.
        """
        return self.variables_types.get(name, None)

    def get_lower_bound(
        self,
        name,  # type: str
    ):  # type:(...) -> ndarray
        """Return the lower bound of a variable.

        Args:
            name: The name of the variable.

        Returns:
            The lower bound of the variable (possibly infinite).
        """
        return self._lower_bounds.get(name)

    def get_upper_bound(
        self,
        name,  # type: str
    ):  # type:(...) -> ndarray
        """Return the upper bound of a variable.

        Args:
            name: The name of the variable.

        Returns:
            The upper bound of the variable (possibly infinite).
        """
        return self._upper_bounds.get(name)

    def get_lower_bounds(
        self,
        variables_names=None,  # type: Optional[Sequence[str]]
    ):  # type: (...) -> ndarray
        """Generate an array of the variables' lower bounds.

        Args:
            variables_names: The names of the variables
                of which the lower bounds are required.
                If None, use the variables of the design space.

        Returns:
            The lower bounds of the variables.
        """
        if self.__norm_data_is_computed and variables_names is None:
            return self.__lower_bounds_array
        return self.dict_to_array(self._lower_bounds, all_var_list=variables_names)

    def get_upper_bounds(
        self,
        variables_names=None,  # type: Optional[Sequence[str]]
    ):  # type: (...) -> ndarray
        """Generate an array of the variables' upper bounds.

        Args:
            variables_names: The names of the variables
                of which the upper bounds are required.
                If None, use the variables of the design space.

        Returns:
            The upper bounds of the variables.
        """
        if self.__norm_data_is_computed and variables_names is None:
            return self.__upper_bounds_array
        return self.dict_to_array(self._upper_bounds, all_var_list=variables_names)

    def set_lower_bound(
        self,
        name,  # type: str
        lower_bound,  # type: ndarray
    ):  # type: (...) -> None
        """Set the lower bound of a variable.

        Args:
            name: The name of the variable.
            lower_bound: The value of the lower bound.

        Raises:
            ValueError: If the variable does not exist.
        """
        if name not in self.variables_names:
            raise ValueError("Variable '{}' is not known.".format(name))

        self._add_bound(name, self.variables_sizes[name], lower_bound, is_lower=True)
        self._add_norm_policy(name)

    def set_upper_bound(
        self,
        name,  # type: str
        upper_bound,  # type: ndarray
    ):  # type: (...) -> None
        """Set the upper bound of a variable.

        Args:
            name: The name of the variable.
            upper_bound: The value of the upper bound.

        Raises:
            ValueError: If the variable does not exist.
        """
        if name not in self.variables_names:
            raise ValueError("Variable '{}' is not known.".format(name))

        self._add_bound(name, self.variables_sizes[name], upper_bound, is_lower=False)
        self._add_norm_policy(name)

    def array_to_dict(
        self,
        x_array,  # type: ndarray
    ):  # type: (...) -> Dict[str,ndarray]
        """Convert the current point into a dictionary indexed by the variables names.

        Args:
            x_array: The current point.

        Returns:
            The dictionary version of the current point.
        """
        x_dict = {}
        current_index = 0
        # order given by self.variables_names
        for name in self.variables_names:
            size = self.variables_sizes[name]
            x_dict[name] = x_array[current_index : current_index + size]
            current_index += size
        return x_dict

    @staticmethod
    def __get_common_dtype(
        x_dict,  # type: Mapping[str,ndarray]
    ):  # type: (...) -> Union[complex128,float64,int32]
        """Return the common data type.

        Use the following rules by parsing the values `x_dict`:

        - there is a complex value: returns `numpy.complex128`,
        - there are real and mixed float/int values: returns `numpy.float64`,
        - there are only integer values: returns `numpy.int32`.

        Args:
            x_dict: The values to be parsed.

        Raises:
            TypeError: If the values of the data dictionary are not NumPy arrays.
        """
        has_float = False
        has_int = False
        for val_arr in x_dict.values():
            if not isinstance(val_arr, ndarray):
                raise TypeError("x_dict values must be ndarray.")
            if val_arr.dtype == complex128:
                return complex128
            if val_arr.dtype == int32:
                has_int = True
            if val_arr.dtype == float64:
                has_float = True
        if has_float:
            return float64
        if has_int:
            return int32
        return float64

    def dict_to_array(
        self,
        x_dict,  # type: Dict[str,ndarray]
        all_vars=True,  # type:bool
        all_var_list=None,  # type: Sequence[str]
    ):  # type: (...) -> ndarray
        """Convert an point as dictionary into an array.

        Args:
            x_dict: The point to be converted.
            all_vars: If True, all the variables to be considered
                shall be in the provided point.
            all_var_list: The variables to be considered.
                If None, use the variables of the design space.

        Returns:
            The point as an array.
        """
        dtype = self.__get_common_dtype(x_dict)
        if all_var_list is None:
            all_var_list = self.variables_names
        if all_vars:
            array_list = [array(x_dict[name], dtype=dtype) for name in all_var_list]
        else:
            array_list = [
                array(x_dict[name], dtype=dtype)
                for name in all_var_list
                if name in x_dict
            ]
        return concatenate(array_list)

    def get_pretty_table(
        self,
        fields=None,  # type: Optional[Sequence[str]]
    ):  # type: (...) -> PrettyTable
        """Build a tabular view of the design space.

        Args:
            fields: The name of the fields to be exported.
                If None, export all the fields.

        Returns:
            A tabular view of the design space.
        """
        if fields is None:
            fields = self.TABLE_NAMES
        table = PrettyTable(fields)
        table.float_format = "%.16g"
        for name in self.variables_names:
            size = self.variables_sizes[name]
            l_b = self._lower_bounds.get(name)
            u_b = self._upper_bounds.get(name)
            var_type = self.variables_types[name]
            curr = self._current_x.get(name)
            for i in range(size):
                data = {
                    "name": name,
                    "value": None,
                    "lower_bound": float("-inf"),
                    "upper_bound": float("inf"),
                    "type": var_type[i],
                }
                if l_b is not None and l_b[i] is not None:
                    data["lower_bound"] = l_b[i]
                if u_b is not None and u_b[i] is not None:
                    data["upper_bound"] = u_b[i]
                if curr is not None:
                    data["value"] = curr[i]
                table.add_row([data[key] for key in fields])

        table.align["name"] = "l"
        table.align["type"] = "l"
        return table

    def export_hdf(
        self,
        file_path,  # type: Union[str,Path]
        append=False,  # type: bool
    ):  # type: (...) -> None
        """Export the design space to an HDF file.

        Args:
            file_path: The path to the file to export the design space.
            append: If True, appends the data in the file.
        """
        mode = "a" if append else "w"

        with h5py.File(file_path, mode) as h5file:
            design_vars_grp = h5file.require_group(self.DESIGN_SPACE_GROUP)
            design_vars_grp.create_dataset(
                self.NAMES_GROUP, data=array(self.variables_names, dtype=string_)
            )

            for name in self.variables_names:
                var_grp = design_vars_grp.require_group(name)
                var_grp.create_dataset(self.SIZE_GROUP, data=self.variables_sizes[name])

                l_b = self._lower_bounds.get(name)
                if l_b is not None:
                    var_grp.create_dataset(self.LB_GROUP, data=l_b)

                u_b = self._upper_bounds.get(name)
                if u_b is not None:
                    var_grp.create_dataset(self.UB_GROUP, data=u_b)

                var_type = self.variables_types[name]
                if var_type is not None:
                    data_array = string_array(var_type)
                    var_grp.create_dataset(
                        self.VAR_TYPE_GROUP, data=data_array, dtype=data_array.dtype
                    )

                value = self._current_x.get(name)
                if value is not None:
                    var_grp.create_dataset(self.VALUE_GROUP, data=self.__to_real(value))

    def import_hdf(
        self,
        file_path,  # type: Union[str,Path]
    ):  # type: (...) -> None
        """Import a design space from an HDF file.

        Args:
            file_path: The path to the file
                containing the description of a design space.
        """
        with h5py.File(file_path, "r") as h5file:
            design_vars_grp = get_hdf5_group(h5file, self.DESIGN_SPACE_GROUP)
            variables_names = get_hdf5_group(design_vars_grp, self.NAMES_GROUP)

            for name in variables_names:
                name = name.decode()
                var_group = get_hdf5_group(design_vars_grp, name)
                l_b = self.__read_opt_attr_array(var_group, self.LB_GROUP)
                u_b = self.__read_opt_attr_array(var_group, self.UB_GROUP)
                var_type = self.__read_opt_attr_array(var_group, self.VAR_TYPE_GROUP)
                value = self.__read_opt_attr_array(var_group, self.VALUE_GROUP)
                size = get_hdf5_group(var_group, self.SIZE_GROUP)[()]
                self.add_variable(name, size, var_type, l_b, u_b, value)

        self.check()

    @staticmethod
    def __read_opt_attr_array(
        var_group,  # type: h5py.Group
        dataset_name,  # type: str
    ):  # type: (...) -> Optional[ndarray]
        """Read data in a group.

        Args:
            var_group: The variable group.
            dataset_name: The name of the dataset.

        Returns:
            The data found in the group, if exist.
            Otherwise, None.
        """
        data = var_group.get(dataset_name)
        if data is not None:
            data = array(data)
        return data

    @staticmethod
    def __to_real(
        data,  # type:ndarray
    ):  # type: (...) -> ndarray
        """Convert complex to real NumPy array.

        Args:
            data: A complex NumPy array.

        Returns:
            A real NumPy array.
        """
        return array(array(data, copy=False).real, dtype=float64)

    def to_complex(self):  # type: (...) -> None
        """Cast the current value to complex."""
        for name, val in self._current_x.items():
            self._current_x[name] = array(val, dtype=complex128)

    def export_to_txt(
        self,
        output_file,  # type: Union[str,Path],
        fields=None,  # type: Optional[Sequence[str]]
        header_char="",  # type: str
        **table_options  # type: Any
    ):  # type: (...) -> None
        """Export the design space to a text file.

        Args:
            output_file: The path to the file.
            fields: The fields to be exported.
                If None, export all fields.
            header_char: The header character.
            **table_options: The names and values of additional attributes
                for the :class:`.PrettyTable` view
                generated by :meth:`get_pretty_table`.
        """
        output_file = Path(output_file)
        table = self.get_pretty_table(fields)
        table.border = False
        for option, val in table_options.items():
            table.__setattr__(option, val)
        with output_file.open("w") as outf:
            table_str = header_char + table.get_string()
            outf.write(table_str)

    @staticmethod
    def read_from_txt(
        input_file,  # type: Union[str,Path]
        header=None,  # type: Optional[Iterable[str]]
    ):  # type: (...) -> DesignSpace
        """Create a design space from a text file.

        Args:
            input_file: The path to the file.
            header: The names of the fields saved in the file.
                If None, read them in the file.

        Returns:
            The design space read from the file.

        Raises:
            ValueError: If the file does not contain the minimal variables
                in its header.
        """
        float_data = genfromtxt(input_file, dtype="float")
        str_data = genfromtxt(input_file, dtype="str")
        if header is None:
            header = strings_to_unicode_list(str_data[0, :].tolist())
            start_read = 1
        else:
            start_read = 0
        if not set(DesignSpace.MINIMAL_FIELDS).issubset(set(header)):
            raise ValueError(
                "Malformed DesignSpace input file {} does not contain "
                "minimal variables in header:"
                "{}; got instead: {}.".format(
                    input_file, DesignSpace.MINIMAL_FIELDS, header
                )
            )
        col_map = {field: i for i, field in enumerate(header)}
        var_names = strings_to_unicode_list(str_data[start_read:, 0].tolist())
        unique_names = []
        prev_name = None
        for name in var_names:  # set([]) does not preserve order !
            if name not in unique_names:
                unique_names.append(name)
                prev_name = name
            elif prev_name != name:
                raise ValueError(
                    "Malformed DesignSpace input file {} contains some variables ({}) "
                    "in a non-consecutive order.".format(input_file, name)
                )

        k = start_read
        design_space = DesignSpace()
        lower_bounds_field = DesignSpace.MINIMAL_FIELDS[1]
        upper_bounds_field = DesignSpace.MINIMAL_FIELDS[2]
        value_field = DesignSpace.TABLE_NAMES[2]
        var_type_field = DesignSpace.TABLE_NAMES[-1]
        for name in unique_names:
            size = var_names.count(name)
            l_b = float_data[k : k + size, col_map[lower_bounds_field]]
            u_b = float_data[k : k + size, col_map[upper_bounds_field]]
            if value_field in col_map:
                value = float_data[k : k + size, col_map[value_field]]
            else:
                value = None
            if var_type_field in col_map:
                var_type = str_data[k : k + size, col_map[var_type_field]].tolist()
            else:
                var_type = DesignVariableType.FLOAT
            design_space.add_variable(name, size, var_type, l_b, u_b, value)
            k += size
        design_space.check()
        return design_space

    def __str__(self):  # type:(...) -> str
        if self.name is None:
            header_suffix = ""
        else:
            header_suffix = " {}".format(self.name)
        header = "Design space:{}\n".format(header_suffix)

        return header + self.get_pretty_table().get_string()

    def project_into_bounds(
        self,
        x_c,  # type: ndarray
        normalized=False,  # type: bool
    ):  # type: (...) -> ndarray
        """Project a vector onto the bounds, using a simple coordinate wise approach.

        Args:
            normalized: If True, then the vector is assumed to be normalized.
            x_c: The vector to be projected onto the bounds.

        Returns:
            The projected vector.
        """
        if not self.__norm_data_is_computed:
            self.__update_normalization_vars()
        if not normalized:
            l_b = self.__lower_bounds_array
            u_b = self.__upper_bounds_array
        else:
            l_b = zeros_like(x_c)
            u_b = ones_like(x_c)
        x_p = array(x_c)
        l_inds = where(x_c < l_b)
        x_p[l_inds] = l_b[l_inds]

        u_inds = where(x_c > u_b)
        x_p[u_inds] = u_b[u_inds]
        return x_p

    def __contains__(
        self,
        variable,  # type: str
    ):  # type: (...) -> bool
        """Return whether the variable is in the design space.

        Args:
            variable: The name of the variable.
        """
        return variable in self.variables_names

    def __len__(self):  # type: (...) -> int
        """The length of the design space, corresponding to the number of variables."""
        return len(self.variables_names)

    def __iter__(self):  # type: (...) -> Iterable[str]
        return iter(self.variables_names)

    def __setitem__(
        self,
        name,  # type: str
        item,  # type: DesignVariable
    ):  # type: (...) -> None
        self.add_variable(
            name,
            size=item.size,
            var_type=item.var_type,
            l_b=item.l_b,
            u_b=item.u_b,
            value=item.value,
        )

    def __eq__(
        self,
        other,  # type: DesignSpace
    ):  # type: (...) -> bool
        if not isinstance(other, self.__class__):
            return False

        if len(other) != len(self):
            return False

        for key, val in self.items():
            if key not in other:
                return False

            hash1 = hash_data_dict(flatten_mapping(val._asdict()))
            hash2 = hash_data_dict(flatten_mapping(other[key]._asdict()))
            if hash1 != hash2:
                return False

        return True

    def __getitem__(
        self,
        name,  # type: str
    ):  # type: (...) -> DesignVariable
        """Return the data associated with a given variable.

        These data are: type, size, lower bound, upper bound and current value.

        Args:
            name: The name of the variable.

        Returns:
            The data associated with the variable.

        Raises:
            ValueError: If the variable name does not exist.
        """
        if name not in self.variables_names:
            raise KeyError("Variable '{}' is not known.".format(name))

        try:
            value = self.get_current_x([name])
        except KeyError:
            value = None

        return DesignVariable(
            size=self.get_size(name),
            var_type=self.get_type(name),
            l_b=self.get_lower_bound(name),
            u_b=self.get_upper_bound(name),
            value=value,
        )

    def extend(
        self,
        other,  # type: DesignSpace
    ):  # type: (...) -> None
        """Extend the design space with another design space.

        Args:
            other: The design space to be appended to the current one.
        """
        for name in other.variables_names:
            size = other.get_size(name)
            var_type = other.get_type(name)
            l_b = other.get_lower_bound(name)
            u_b = other.get_upper_bound(name)
            value = other.get_current_x_dict()[name]
            self.add_variable(name, size, var_type, l_b, u_b, value)

    @staticmethod
    def __cast_array_to_list(
        value,  # type: Union[str,int,ndarray]
    ):  # type: (...) -> Union[str,int,List[Union[str,int]]]
        """Convert a value to a ``List`` if it is a NumPy array.

        Args:
            value: The value to be casted.

        Returns:
            Either the original value or the NumPy array converted to a ``List``.
        """
        return value if not isinstance(value, ndarray) else value.tolist()

    @classmethod
    def __cast_mapping(
        cls,
        mapping,  # type: Mapping[str,Union[str,int,ndarray]]
    ):  # type: (...) -> Dict[str,Union[str,int,List[Union[str,int]]]]
        """Convert the NumPy arrays of a mapping to ``List``.

        Args:
            mapping: The value to be casted.

        Returns:
            The original mapping with NumPy values converted to a ``List``.
        """
        return {
            key: {
                sub_key: cls.__cast_array_to_list(sub_val)
                for sub_key, sub_val in val.items()
            }
            for key, val in mapping.items()
        }



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or
#                      initial documentation
#        :author:  Francois Gallard
#    OTHER AUTHORS   - MACROSCOPIC CHANGES

"""Excel based discipline."""

from __future__ import division, unicode_literals

import atexit
import os
import shutil
import tempfile
from typing import Any, List, Mapping, Optional
from uuid import uuid4

from numpy import array

from gemseo.core.discipline import MDODiscipline
from gemseo.utils.py23_compat import Path

cwd = Path.cwd()
try:
    import xlwings
except ImportError:
    # error will be reported if the discipline is used
    os.chdir(str(cwd))
    xlwings = None


class XLSDiscipline(MDODiscipline):
    """Wraps an excel workbook into a discipline.

    .. warning::

        As this wrapper relies on the `xlswings library <https://www.xlwings.org>`__
        to handle macros and interprocess communication, it
        is only working under Windows and MacOS.
    """

    _ATTR_TO_SERIALIZE = MDODiscipline._ATTR_TO_SERIALIZE + (
        "_xls_file_path",
        "input_names",
        "output_names",
        "macro_name",
        "_copy_xls_at_setstate",
    )

    def __init__(
        self,
        xls_file_path,  # type: str
        macro_name="execute",  # type: Optional[str]
        copy_xls_at_setstate=False,  # type: bool
    ):  # type: (...) -> None
        """Initialize xls file path and macro.

        Inputs must be specified in the "Inputs" sheet, in the following
        format (A and B are the first two columns).

        +---+---+
        | A | B |
        +===+===+
        | a | 1 |
        +---+---+
        | b | 2 |
        +---+---+

        Where a is the name of the first input, and 1 is its default value
        b is the name of the second one, and 2 its default value.

        There must be no empty lines between the inputs.


           The number of rows is arbitrary but they must be contiguous and
           start at line 1.

           The same applies for the "Outputs" sheet
           (A and B are the first two columns).


        +---+---+
        | A | B |
        +===+===+
        | c | 3 |
        +---+---+

        Where c is the only output. They may be multiple.

           If the file is a .xlsm, a macro named "execute" *must* exist
           and will be called by the _run method before retrieving
           the outputs. The macro has no arguments, it takes its inputs in the
           Inputs sheet, and write the outputs to the "Outputs" sheet.

           Alternatively, the user may provide a macro name to the constructor,
           or None if no macro shall be executed.

        Args:
            xls_file_path: The path to the excel file. If the file is a `.xlsm`,
                a macro named "execute" must exist and will be called by the _run
                method before retrieving the outputs.
            macro_name: The name of the macro to be executed for a `.xlsm` file.
                If None is provided, do not run a macro.
            copy_xls_at_setstate: If True, create a copy of the original Excel file
                for each of the pickled parallel processes. This option is required
                to be set to True for parallelization in Windows platforms.

        Raises:
            ImportError: If `xlwings` cannot be imported.
        """
        if xlwings is None:
            raise ImportError("cannot import xlwings")
        super(XLSDiscipline, self).__init__()
        self._xls_file_path = Path(xls_file_path)
        self._xls_app = None
        self.macro_name = macro_name
        self.input_names = None
        self.output_names = None
        self._book = None
        self._copy_xls_at_setstate = copy_xls_at_setstate

        self.__init_workbook()
        self._init_grammars()
        self._init_defaults()
        self.re_exec_policy = self.RE_EXECUTE_DONE_POLICY

    def __init_workbook(self):  # type: (...) -> None
        """Initialize a workbook.

        Raises:
            ValueError: If there is no sheet in the Excel file
                named "Inputs" or if there is no sheet named
                "Outputs".
        """

        try:
            self._xls_app = xlwings.App(visible=False)
            self._xls_app.interactive = False
        # wide except because I cannot tell what is the exception raised by xlwings
        except:  # noqa: E722,B001
            raise RuntimeError("xlwings requires Microsoft Excel")

        # Close the app when exiting
        atexit.register(self._xls_app.quit)

        self._book = self._xls_app.books.open(str(self._xls_file_path))
        sh_names = [sheet.name for sheet in self._book.sheets]

        if "Inputs" not in sh_names:
            raise ValueError(
                "Workbook must contain a sheet named 'Inputs' "
                "that define the inputs of the discipline"
            )
        if "Outputs" not in sh_names:
            raise ValueError(
                "Workbook must contain a sheet named 'Outputs' "
                "that define the outputs of the discipline"
            )

    def __setstate__(
        self, state  # type: Mapping[str, Any]
    ):  # type: (...) -> None
        super(XLSDiscipline, self).__setstate__(state)
        self._book = None
        self._xls_app = None
        if self._copy_xls_at_setstate:
            temp_dir = Path(tempfile.gettempdir())
            temp_path = temp_dir / self._xls_file_path.name.replace(
                ".xls", str(uuid4()) + ".xls"
            )
            shutil.copy2(str(self._xls_file_path), str(temp_path))
            self._xls_file_path = temp_path
        self.__init_workbook()

    def get_attributes_to_serialize(self):  # type: (...) -> List[str]
        """Overload pickle method to define which attributes are to be serialized.

        Returns:
             The attributes to serialize.
        """
        base = super(XLSDiscipline, self).get_attributes_to_serialize()
        base += [
            "_xls_file_path",
            "input_names",
            "output_names",
            "macro_name",
            "_copy_xls_at_setstate",
        ]
        return base

    def __read_sheet_col(
        self,
        sheet_name,  # type: str
        column=0,  # type: int
    ):  # type: (...) -> List[List[str], List[str], List[float], List[float]]
        """Read a specific column of the sheet.

        Args:
            sheet_name: The name of the sheet to be read.
            column: The number of the column to be read.

        Returns:
            The column values.
        """
        sht = self._book.sheets[sheet_name]
        i = 0
        value = sht[i, column].value
        values = []
        while value is not None:
            values.append(value)
            i += 1
            value = sht[i, column].value
        return values

    def _init_grammars(self):  # type: (...) -> None
        """Initialize grammars by parsing the Inputs and Outputs sheets."""
        self.input_names = self.__read_sheet_col("Inputs", 0)
        self.output_names = self.__read_sheet_col("Outputs", 0)
        self.input_grammar.initialize_from_data_names(self.input_names)
        self.output_grammar.initialize_from_data_names(self.output_names)

    def _init_defaults(self):  # type: (...) -> None
        """Initialize the default input values.

        Raises:
            ValueError: If the "Inputs" sheet does not have the same number of
                entries in the name column and the value column.
        """
        inputs = self.__read_sheet_col("Inputs", 1)
        if len(inputs) != len(self.input_names):
            msg = (
                "Inconsistent Inputs sheet, names (first columns) and"
                " values column (second) must be of the same length"
            )
            raise ValueError(msg)

        self.default_inputs = {k: array([v]) for k, v in zip(self.input_names, inputs)}

    def __write_inputs(
        self, input_data  # type: Mapping[str, float]
    ):  # type: (...) -> None
        """Write the inputs values to the Inputs sheet."""
        sht = self._book.sheets["Inputs"]
        for i, key in enumerate(self.input_names):
            sht[i, 1].value = input_data[key][0]

    def _run(self):  # type: (...) -> None
        """Run the discipline.

        Eventually calls the execute macro.

        Raises:
            RuntimeError: If the macro fails to be executed.
            ValueError: If the "Outputs" sheet does not have the same number of
                entries in the name column and the value column.
        """
        self.__write_inputs(self.local_data)
        if self._xls_file_path.match("*.xlsm") and self.macro_name is not None:
            try:
                self._xls_app.api.Application.Run(self.macro_name)
            except Exception as err:
                macro_name = self.macro_name
                msg = "Failed to run '{}' macro: {}.".format(macro_name, err)
                raise RuntimeError(msg)
        out_vals = self.__read_sheet_col("Outputs", 1)
        if len(out_vals) != len(self.output_names):
            msg = (
                "Inconsistent Outputs sheet, names (first columns) and "
                "values column (second) must be of the same length."
            )
            raise ValueError(msg)

        outputs = {k: array([v]) for k, v in zip(self.output_names, out_vals)}
        self.store_local_data(**outputs)



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - API and implementation and/or documentation
#       :author: Damien Guenot - 26 avr. 2016
#       :author: Francois Gallard, refactoring
#    OTHER AUTHORS   - MACROSCOPIC CHANGES

"""
Driver library
==============

A driver library aims to solve an :class:`.OptimizationProblem`
using a particular algorithm from a particular family of numerical methods.
This algorithm will be in charge of evaluating the objective and constraints
functions at different points of the design space, using the
:meth:`.DriverLib.execute` method.
The most famous kinds of numerical methods to solve an optimization problem
are optimization algorithms and design of experiments (DOE). A DOE driver
browses the design space agnostically, i.e. without taking into
account the function evaluations. On the contrary, an optimization algorithm
uses this information to make the journey through design space
as relevant as possible in order to reach as soon as possible the optimum.
These families are implemented in :class:`.DOELibrary`
and :class:`.OptimizationLibrary`.
"""

from __future__ import division, unicode_literals

import io
import logging
import string
from time import time
from typing import Callable, List, Optional, Union

import tqdm
from numpy import ndarray, ones_like, where, zeros_like
from tqdm.utils import _unicode, disp_len

from gemseo.algos.algo_lib import AlgoLib
from gemseo.algos.opt_problem import OptimizationProblem
from gemseo.algos.opt_result import OptimizationResult
from gemseo.algos.stop_criteria import (
    DesvarIsNan,
    FtolReached,
    FunctionIsNan,
    MaxIterReachedException,
    MaxTimeReached,
    TerminationCriterion,
    XtolReached,
)

DriverLibOptionType = Union[str, float, int, bool, List[str], ndarray]
LOGGER = logging.getLogger(__name__)


class TqdmToLogger(io.StringIO):
    """Redirect tqdm output to the gemseo logger."""

    def write(self, buf):
        """Write buffer."""
        buf = buf.strip(string.whitespace)
        if buf:
            LOGGER.info(buf)


class ProgressBar(tqdm.tqdm):
    """Extend tqdm progress bar with better time units.

    Use hour, day or week for slower processes.
    """

    @classmethod
    def format_meter(cls, n, total, elapsed, **kwargs):
        if elapsed != 0.0:
            rate, unit = cls.__convert_rate(total, elapsed)
            kwargs["rate"] = rate
            kwargs["unit"] = unit
        meter = tqdm.tqdm.format_meter(n, total, elapsed, **kwargs)
        # remove the unit suffix that is hard coded in tqdm
        return meter.replace("/s,", ",").replace("/s]", "]")

    @staticmethod
    def __convert_rate(total, elapsed):
        rps = float(total) / elapsed
        if rps >= 0:
            rate = rps
            unit = "sec"

        rpm = rps * 60.0
        if rpm < 60.0:
            rate = rpm
            unit = "min"

        rph = rpm * 60.0
        if rph < 60.0:
            rate = rph
            unit = "hour"

        rpd = rph * 24.0
        if rpd < 24.0:
            rate = rpd
            unit = "day"

        return rate, " it/{}".format(unit)

    def status_printer(
        self, file  # type: Union[io.TextIOWrapper, io.StringIO]
    ):  # type: (...) -> Callable[[str], None]
        """Overload the status_printer method to avoid the use of closures.

        Args:
            file: Specifies where to output the progress messages.

        Returns:
            The function to print the status in the progress bar.
        """
        self._last_len = [0]
        return self._print_status

    def _print_status(self, s):
        len_s = disp_len(s)
        self.fp.write(
            _unicode("\r{}{}".format(s, (" " * max(self._last_len[0] - len_s, 0))))
        )
        self.fp.flush()
        self._last_len[0] = len_s

    def __getstate__(self):
        state = self.__dict__.copy()
        # A file-like stream cannot be pickled.
        del state["fp"]
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        # Set back the file-like stream to its state as done in tqdm.__init__.
        self.fp = tqdm.utils.DisableOnWriteError(TqdmToLogger(), tqdm_instance=self)


class DriverLib(AlgoLib):
    """Abstract class for DOE & optimization libraries interfaces.

    Lists available methods in the library for the proposed
    problem to be solved.

    To integrate an optimization package, inherit from this class
    and put your file in gemseo.algos.doe or gemseo.algo.opt packages.
    """

    USER_DEFINED_GRADIENT = OptimizationProblem.USER_GRAD
    COMPLEX_STEP_METHOD = OptimizationProblem.COMPLEX_STEP
    FINITE_DIFF_METHOD = OptimizationProblem.FINITE_DIFFERENCES

    DIFFERENTIATION_METHODS = [
        USER_DEFINED_GRADIENT,
        COMPLEX_STEP_METHOD,
        FINITE_DIFF_METHOD,
    ]

    REQUIRE_GRAD = "require_grad"
    HANDLE_EQ_CONS = "handle_equality_constraints"
    HANDLE_INEQ_CONS = "handle_inequality_constraints"
    POSITIVE_CONSTRAINTS = "positive_constraints"
    INEQ_TOLERANCE = "ineq_tolerance"
    EQ_TOLERANCE = "eq_tolerance"
    MAX_TIME = "max_time"
    USE_DATABASE_OPTION = "use_database"
    NORMALIZE_DESIGN_SPACE_OPTION = "normalize_design_space"
    ROUND_INTS_OPTION = "round_ints"
    WEBSITE = "website"
    DESCRIPTION = "description"
    MAX_DS_SIZE_PRINT = 40

    def __init__(self):
        """Constructor."""
        # Library settings and check
        super(DriverLib, self).__init__()
        self.__progress_bar = None
        self.__max_iter = 0
        self.__iter = 0
        self._start_time = None
        self._max_time = None
        self.__message = None

    def deactivate_progress_bar(self):  # type: (...) -> None
        """Deactivate the progress bar."""
        self.__progress_bar = None

    def init_iter_observer(
        self,
        max_iter,  # type: int
        message,  # type: str
    ):  # type: (...) -> None
        """Initialize the iteration observer.

        It will handle the stopping criterion and the logging of the progress bar.

        Args:
            max_iter: The maximum number of iterations.
            message: The message to display at the beginning.

        Raises:
            ValueError: If the `max_iter` is not greater than or equal to one.
        """
        if max_iter < 1:
            raise ValueError("max_iter must be >=1, got {}".format(max_iter))
        self.__max_iter = max_iter
        self.__iter = 0
        self.__message = message
        self.__progress_bar = ProgressBar(
            total=self.__max_iter,
            desc=self.__message,
            ascii=False,
            file=TqdmToLogger(),
        )
        self._start_time = time()
        self.problem.max_iter = max_iter

    def __set_progress_bar_objective_value(
        self, x_vect  # type: ndarray
    ):  # type: (...) -> None
        """Set the objective value in the progress bar.

        Args:
            x_vect: The design variables values.
        """
        value = self.problem.database.get_f_of_x(self.problem.objective.name, x_vect)

        if value is not None:
            # if maximization problem: take the opposite
            if not self.problem.minimize_objective:
                value = -value

            self.__progress_bar.set_postfix(refresh=False, obj=value)
            self.__progress_bar.update()
        else:
            self.__progress_bar.update()

    def new_iteration_callback(
        self, x_vect=None  # type: Optional[ndarray]
    ):  # type: (...) -> None
        """Callback called at each new iteration, i.e. every time a design vector that
        is not already in the database is proposed by the optimizer.

        Iterate the progress bar, implement the stop criteria.

        Args:
            x_vect: The design variables values. If None, use the values of the
                last iteration.

        Raises:
            MaxTimeReached: If the elapsed time is greater than the maximum
                execution time.
        """
        # First check if the max_iter is reached and update the progress bar
        self.__iter += 1
        self.problem.current_iter = self.__iter

        if self._max_time > 0:
            delta_t = time() - self._start_time
            if delta_t > self._max_time:
                raise MaxTimeReached()

        if self.__progress_bar is not None:
            if x_vect is None:
                x_vect = self.problem.database.get_x_by_iter(-1)
            self.__set_progress_bar_objective_value(x_vect)

    def finalize_iter_observer(self):  # type: (...) -> None
        """Finalize the iteration observer."""
        if self.__progress_bar is not None:
            self.__progress_bar.close()

    def _pre_run(
        self,
        problem,  # type: OptimizationProblem
        algo_name,  # type: str
        **options  # type: DriverLibOptionType
    ):  # type: (...) -> None
        """To be overridden by subclasses. Specific method to be executed just before
        _run method call.

        Args:
            problem: The optimization problem.
            algo_name: The name of the algorithm.
            **options: The options of the algorithm,
                see the associated JSON file.
        """
        self._max_time = options.get(self.MAX_TIME, 0.0)

    def _post_run(self, problem, algo_name, result, **options):  # pylint: disable=W0613
        """To be overridden by subclasses Specific method to be executed just after _run
        method call.

        :param problem: the problem to be solved
        :param algo_name: name of the algorithm
        :param result: result of the run such as an OptimizationResult
        :param options: the options dict for the algorithm, see associated JSON file
        """
        LOGGER.info("%s", result)
        problem.solution = result
        if result.x_opt is not None:
            problem.design_space.set_current_x(result)
        if problem.design_space.dimension <= self.MAX_DS_SIZE_PRINT:
            LOGGER.info("%s", problem.design_space)

    def execute(self, problem, algo_name=None, **options):
        """Executes the driver.

        :param problem: the problem to be solved
        :param algo_name: name of the algorithm
            if None, use self.algo_name
            which may have been set by the factory (Default value = None)
        :param options: the options dict for the algorithm
        """
        self.problem = problem
        if algo_name is not None:
            self.algo_name = algo_name

        if self.algo_name is None:
            raise ValueError(
                "Algorithm name must be either passed as "
                "argument or set by the attribute 'algo_name'."
            )

        self._check_algorithm(self.algo_name, problem)
        options = self._update_algorithm_options(**options)
        self.internal_algo_name = self.lib_dict[self.algo_name][self.INTERNAL_NAME]

        problem.check()
        problem.preprocess_functions(
            normalize=options.get(self.NORMALIZE_DESIGN_SPACE_OPTION, True),
            use_database=options.get(self.USE_DATABASE_OPTION, True),
            round_ints=options.get(self.ROUND_INTS_OPTION, True),
        )

        try:  # Term criteria such as max iter or max_time can be triggered in pre_run
            self._pre_run(problem, self.algo_name, **options)
            result = self._run(**options)
        except TerminationCriterion as error:
            result = self._termination_criterion_raised(error)
        self.finalize_iter_observer()
        self.problem.clear_listeners()
        self._post_run(problem, algo_name, result, **options)

        return result

    def _process_specific_option(self, options, option_key):
        """Process one option as a special treatment, at the begining of the general
        treatment and checks of _process_options.

        Args:
            options: The options as preprocessed by _process_options.
            option_key: The current option key to process.
        """
        if option_key == self.INEQ_TOLERANCE:
            self.problem.ineq_tolerance = options[option_key]
            del options[option_key]
        elif option_key == self.EQ_TOLERANCE:
            self.problem.eq_tolerance = options[option_key]
            del options[option_key]

    def _termination_criterion_raised(self, error):  # pylint: disable=W0613
        """Retrieve the best known iterate when max iter has been reached.

        :param error: the obtained error from the algorithm
        """
        if isinstance(error, TerminationCriterion):
            message = ""
            if isinstance(error, MaxIterReachedException):
                message = "Maximum number of iterations reached."
            elif isinstance(error, FunctionIsNan):
                message = "Function value or gradient or constraint is NaN, "
                message += "and problem.stop_if_nan is set to True."
            elif isinstance(error, DesvarIsNan):
                message = "Design variables are NaN."
            elif isinstance(error, XtolReached):
                message = "Successive iterates of the design variables "
                message += "are closer than xtol_rel or xtol_abs."
            elif isinstance(error, FtolReached):
                message = "Successive iterates of the objective function "
                message += "are closer than ftol_rel or ftol_abs."
            elif isinstance(error, MaxTimeReached):
                message = "Maximum time reached : {} seconds.".format(self._max_time)
            message += " GEMSEO Stopped the driver"
        else:
            message = error.args[0]

        result = self.get_optimum_from_database(message)
        return result

    def get_optimum_from_database(self, message=None, status=None):
        """Retrieves the optimum from the database and builds an optimization result
        object from it.

        :param message: Default value = None)
        :param status: Default value = None)
        """
        problem = self.problem
        if len(problem.database) == 0:
            return OptimizationResult(
                optimizer_name=self.algo_name,
                message=message,
                status=status,
                n_obj_call=0,
            )
        x_0 = problem.database.get_x_by_iter(0)
        # compute the best feasible or infeasible point
        f_opt, x_opt, is_feas, c_opt, c_opt_grad = problem.get_optimum()
        if f_opt is not None and not problem.minimize_objective:
            f_opt = -f_opt
        return OptimizationResult(
            x_0=x_0,
            x_opt=x_opt,
            f_opt=f_opt,
            optimizer_name=self.algo_name,
            message=message,
            status=status,
            n_obj_call=problem.objective.n_calls,
            is_feasible=is_feas,
            constraints_values=c_opt,
            constraints_grad=c_opt_grad,
        )

    def _get_options(self, **options):
        """Retrieves the options of the library To be overloaded by subclasses Used to
        define default values for options using keyword arguments.

        :param options: options of the driver
        """
        raise NotImplementedError()

    def _run(self, **options):
        """Runs the algorithm, to be overloaded by subclasses.

        :param options: the options dict for the algorithm
        """
        raise NotImplementedError()

    def is_algo_requires_grad(self, algo_name):
        """Returns True if the algorithm requires a gradient evaluation.

        :param algo_name: name of the algorithm
        """
        lib_alg = self.lib_dict.get(algo_name, None)
        if lib_alg is None:
            raise ValueError("Algorithm {} is not available.".format(algo_name))

        return lib_alg.get(self.REQUIRE_GRAD, False)

    def get_x0_and_bounds_vects(self, normalize_ds):
        """Gets x0, bounds, normalized or not depending on algo options, all as numpy
        arrays.

        :param normalize_ds: if True, normalizes all input vars
               that are not integers, according to design space
               normalization policy
        :returns: x, lower bounds, upper bounds
        """
        dspace = self.problem.design_space
        l_b = dspace.get_lower_bounds()
        u_b = dspace.get_upper_bounds()

        # remove normalization from options for algo
        if normalize_ds:
            norm_dict = dspace.normalize
            norm_array = dspace.dict_to_array(norm_dict)
            xvec = self.problem.get_x0_normalized()
            l_b = where(norm_array, zeros_like(xvec), l_b)
            u_b = where(norm_array, ones_like(xvec), u_b)
        else:
            xvec = self.problem.design_space.get_current_x()

        return xvec, l_b, u_b

    def ensure_bounds(self, orig_func, normalize=True):
        """Project the design vector onto the design space before execution.

        :param orig_func: the original function
        :param normalize: if True, use the normalized design space
        :returns: the wrapped function
        """

        def wrapped_func(x_vect):
            x_proj = self.problem.design_space.project_into_bounds(x_vect, normalize)
            return orig_func(x_proj)

        return wrapped_func



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                         documentation
#        :author: Syver Doving Agdestein
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
r"""The polynomial model for regression.

Polynomial regression class is a particular case of the linear regression,
where the input data is transformed before the regression is applied.
This transform consists of creating a matrix of monomials (Vandermonde)
by raising the input data to different powers up to a certain degree :math:`D`.
In the case where there is only one input variable,
the input data :math:`(x_i)_{i=1, \dots, n}\in\mathbb{R}^n` is transformed
into the Vandermonde matrix

.. math::

    \begin{pmatrix}
        x_1^1  & x_1^2  & \cdots & x_1^D\\
        x_2^1  & x_2^2  & \cdots & x_2^D\\
        \vdots & \vdots & \ddots & \vdots\\
        x_n^1  & x_n^2  & \cdots & x_n^D\\
    \end{pmatrix}
    = (x_i^d)_{i=1, \dots, n;\ d=1, \dots, D}.

The output is expressed as a weighted sum of monomials:

.. math::

     y = w_0 + w_1 x^1 + w_2 x^2 + ... + w_D x^D,

where the coefficients :math:`(w_1, w_2, ..., w_d)` and the intercept :math:`w_0`
are estimated by least square regression.

In the case of a multidimensional input,
i.e. :math:`X = (x_{ij})_{i=1,\dots,n; j=1,\dots,m}`,
where :math:`n` is the number of samples and :math:`m` is the number of input variables,
the Vandermonde matrix is expressed
through different combinations of monomials of degree :math:`d, (1 \leq d \leq D)`;
e.g. for three variables :math:`(x, y, z)` and degree :math:`D=3`,
the different terms are
:math:`x`, :math:`y`, :math:`z`, :math:`x^2`, :math:`xy`, :math:`xz`,
:math:`y^2`, :math:`yz`, :math:`z^2`, :math:`x^3`, :math:`x^2y` etc.
More generally,
for m input variables,
the total number of monomials of degree :math:`1 \leq d \leq D` is given
by :math:`P = \binom{m+D}{m} = \frac{(m+D)!}{m!D!}`.
In the case of 3 input variables given above,
the total number of monomial combinations of degree lesser than or equal to three
is thus :math:`P = \binom{6}{3} = 20`.
The linear regression has to identify the coefficients :math:`(w_1, \dots, w_P)`,
in addition to the intercept :math:`w_0`.

This concept is implemented through the :class:`.PolynomialRegression` class
which inherits from the :class:`.MLRegressionAlgo` class.

Dependence
----------
The polynomial regression model relies on the LinearRegression class
of the `LinearRegression <https://scikit-learn.org/stable/modules/
linear_model.html>`_ and  `PolynomialFeatures <https://scikit-learn.org/stable/
modules/generated/sklearn.preprocessing.PolynomialFeatures.html>`_ classes of
the `scikit-learn library <https://scikit-learn.org/stable/modules/
linear_model.html>`_.
"""
from __future__ import division, unicode_literals

import logging
import pickle
from typing import Iterable, Optional, Union

from numpy import concatenate, ndarray, where, zeros
from sklearn.preprocessing import PolynomialFeatures

from gemseo.core.dataset import Dataset
from gemseo.mlearning.core.ml_algo import DataType, TransformerType
from gemseo.mlearning.regression.linreg import LinearRegression
from gemseo.utils.py23_compat import Path

LOGGER = logging.getLogger(__name__)


class PolynomialRegression(LinearRegression):
    """Polynomial regression."""

    LIBRARY = "scikit-learn"
    ABBR = "PolyReg"

    def __init__(
        self,
        data,  # type: Dataset
        degree,  # type: int
        transformer=None,  # type: Optional[TransformerType]
        input_names=None,  # type: Optional[Iterable[str]]
        output_names=None,  # type: Optional[Iterable[str]]
        fit_intercept=True,  # type: bool
        penalty_level=0.0,  # type: float
        l2_penalty_ratio=1.0,  # type: float
        **parameters  # type: Optional[Union[float,int,str,bool]]
    ):  # type: (...) -> None
        """
        Args:
            degree: The polynomial degree.
            fit_intercept: Whether to fit the intercept.
            penalty_level: The penalty level greater or equal to 0.
                If 0, there is no penalty.
            l2_penalty_ratio: The penalty ratio
                related to the l2 regularization.
                If 1, the penalty is the Ridge penalty.
                If 0, this is the Lasso penalty.
                Between 0 and 1, the penalty is the ElasticNet penalty.

        Raises:
            ValueError: If the degree is lower than one.
        """
        super(PolynomialRegression, self).__init__(
            data,
            degree=degree,
            transformer=transformer,
            input_names=input_names,
            output_names=output_names,
            fit_intercept=fit_intercept,
            penalty_level=penalty_level,
            l2_penalty_ratio=l2_penalty_ratio,
            **parameters
        )
        self._poly = PolynomialFeatures(degree=degree, include_bias=False)
        self.parameters["degree"] = degree
        if degree < 1:
            raise ValueError("Degree must be >= 1.")

    def _fit(
        self,
        input_data,  # type: ndarray
        output_data,  # type: ndarray
    ):  # type: (...) -> None
        input_data = self._poly.fit_transform(input_data)
        super(PolynomialRegression, self)._fit(input_data, output_data)

    def _predict(
        self,
        input_data,  # type: ndarray
    ):  # type: (...) -> ndarray
        input_data = self._poly.transform(input_data)
        return super(PolynomialRegression, self)._predict(input_data)

    def _predict_jacobian(
        self,
        input_data,  # type: ndarray
    ):  # type: (...) -> ndarray
        # Dimensions:
        # powers:        (           ,            ,  n_powers ,  n_inputs )
        # coefs:         (           ,  n_outputs ,  n_powers ,           )
        # jac_coefs:     (           ,  n_outputs ,  n_powers ,  n_inputs )
        # vandermonde:   ( n_samples ,            ,  n_powers ,           )
        # contributions: ( n_samples ,  n_outputs ,  n_powers ,  n_inputs )
        # jacobians:     ( n_samples ,  n_outputs ,           ,  n_inputs )
        #
        # n_powers is given by the formula
        # n_powers = binom(n_inputs+degree, n_inputs)+1

        vandermonde = self._poly.transform(input_data)

        powers = self._poly.powers_
        n_inputs = self._poly.n_input_features_
        n_powers = self._poly.n_output_features_
        n_outputs = self.algo.coef_.shape[0]
        coefs = self.get_coefficients(False)

        jac_intercept = zeros((n_outputs, n_inputs))
        jac_coefs = zeros((n_outputs, n_powers, n_inputs))

        # Compute partial derivatives with respect to each input separately
        for index in range(n_inputs):

            # Coefficients of monomial derivatives
            dcoefs = powers[None, :, index] * coefs

            # Powers of monomial derivatives
            dpowers = powers.copy()
            dpowers[:, index] -= 1

            # Keep indices of remaining monomials only
            mask_zero = (dpowers == 0).prod(axis=1) == 1
            mask_keep = dpowers[:, index] >= 0
            mask_keep[mask_zero == 1] = False

            # Extract intercept for Jacobian (0th order term)
            dintercept = dcoefs[:, mask_zero].flatten()

            # Filter kept terms
            dcoefs = dcoefs[:, mask_keep]  # Coefficients of kept terms
            dpowers = dpowers[mask_keep]  # Power keys of kept terms

            # Find indices for the given powers
            inds_keep = [
                where((powers == dpowers[i]).prod(axis=1) == 1)
                for i in range(dpowers.shape[0])
            ]
            if len(inds_keep) > 0:
                inds_keep = concatenate(inds_keep).flatten()

            # Coefficients of partial derivatives in terms of original powers
            jac_intercept[:, index] = dintercept
            jac_coefs[:, inds_keep, index] = dcoefs

        # Assemble polynomial (sum of weighted monomials)
        contributions = jac_coefs[None] * vandermonde[:, None, :, None]
        jacobians = jac_intercept + contributions.sum(axis=2)

        return jacobians

    def get_coefficients(
        self,
        as_dict=False,  # type:bool
    ):  # type: (...) -> DataType
        """Return the regression coefficients of the linear model.

        Args:
            as_dict: If True, return the coefficients as a dictionary of Numpy arrays
                indexed by the names of the coefficients.
                Otherwise, return the coefficients as a Numpy array.
                For now the only valid value is False.

        Returns:
            The regression coefficients of the linear model.

        Raises:
            NotImplementedError: If the coefficients are required as a dictionary.
        """
        coefficients = self.coefficients
        if as_dict:
            raise NotImplementedError(
                "For now the coefficients can only be obtained "
                "in the form of a NumPy array"
            )
        return coefficients

    def _save_algo(
        self,
        directory,  # type: Path
    ):  # type: (...) -> None
        super(PolynomialRegression, self)._save_algo(directory)
        with (directory / "poly.pkl").open("wb") as handle:
            pickle.dump(self._poly, handle)

    def load_algo(
        self,
        directory,  # type: Union[str,Path]
    ):  # type: (...) -> None
        directory = Path(directory)
        super(PolynomialRegression, self).load_algo(directory)
        with (directory / "poly.pkl").open("rb") as handle:
            poly = pickle.load(handle)
        self._poly = poly



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - API and implementation and/or documentation
#        :author: Charlie Vanaret
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""An advanced MDA splitting algorithm based on graphs."""
from __future__ import division, unicode_literals

import logging
from itertools import repeat
from multiprocessing import cpu_count
from os.path import join, split
from typing import Any, Iterable, List, Mapping, Optional, Sequence, Tuple, Union

from gemseo.api import create_mda
from gemseo.core.chain import MDOChain
from gemseo.core.coupling_structure import MDOCouplingStructure
from gemseo.core.discipline import MDODiscipline
from gemseo.core.execution_sequence import SerialExecSequence
from gemseo.mda.mda import MDA

LOGGER = logging.getLogger(__name__)
N_CPUS = cpu_count()


class MDAChain(MDA):
    """A chain of sub-MDAs.

    The execution sequence is provided by the :class:`.DependencyGraph`.
    """

    _ATTR_TO_SERIALIZE = MDA._ATTR_TO_SERIALIZE + (
        "mdo_chain",
        "_chain_linearize",
        "lin_cache_tol_fact",
        "assembly",
        "coupling_structure",
        "linear_solver",
        "linear_solver_options",
        "linear_solver_tolerance",
        "matrix_type",
        "use_lu_fact",
        "all_couplings",
    )

    def __init__(
        self,
        disciplines,  # type: Sequence[MDODiscipline]
        sub_mda_class="MDAJacobi",  # type: str
        max_mda_iter=20,  # type: int
        name=None,  # type: Optional[str]
        n_processes=N_CPUS,  # type: int
        chain_linearize=False,  # type: bool
        tolerance=1e-6,  # type: float
        linear_solver_tolerance=1e-12,  # type: float
        use_lu_fact=False,  # type: bool
        grammar_type=MDODiscipline.JSON_GRAMMAR_TYPE,  # type: str
        coupling_structure=None,  # type: Optional[MDOCouplingStructure]
        sub_coupling_structures=None,  # type: Optional[Iterable[MDOCouplingStructure]]
        log_convergence=False,  # type: bool
        linear_solver="DEFAULT",  # type: str
        linear_solver_options=None,  # type: Mapping[str,Any]
        **sub_mda_options  # type: Optional[Union[float, int, bool, str]]
    ):
        """
        Args:
            sub_mda_class: The class name of the sub-MDA.
            n_processes: The number of processes.
            chain_linearize: Whether to linearize the chain of execution.
                Otherwise, linearize the overall MDA with base class method.
                This last option is preferred to minimize computations in adjoint mode,
                while in direct mode, linearizing the chain may be cheaper.
            sub_coupling_structures: The coupling structures to be used by the sub-MDAs.
                If None, they are created from the sub-disciplines.
            **sub_mda_options: The options to be passed to the sub-MDAs.
        """
        self.n_processes = n_processes
        self.mdo_chain = None
        self._chain_linearize = chain_linearize
        self.sub_mda_list = []

        # compute execution sequence of the disciplines
        super(MDAChain, self).__init__(
            disciplines,
            max_mda_iter=max_mda_iter,
            name=name,
            tolerance=tolerance,
            linear_solver_tolerance=linear_solver_tolerance,
            use_lu_fact=use_lu_fact,
            grammar_type=grammar_type,
            coupling_structure=coupling_structure,
            linear_solver=linear_solver,
            linear_solver_options=linear_solver_options,
        )

        if (
            not self.coupling_structure.get_all_couplings()
            and not self._chain_linearize
        ):
            LOGGER.warning("No coupling in MDA, switching chain_linearize to True")
            self._chain_linearize = True

        self._create_mdo_chain(
            disciplines,
            sub_mda_class=sub_mda_class,
            sub_coupling_structures=sub_coupling_structures,
            **sub_mda_options
        )

        self.log_convergence = log_convergence

        self._initialize_grammars()
        self._check_consistency()
        self._set_default_inputs()
        self._compute_input_couplings()

        # cascade the tolerance
        for sub_mda in self.sub_mda_list:
            sub_mda.tolerance = self.tolerance

    @MDA.log_convergence.setter
    def log_convergence(
        self,
        value,  # type: bool
    ):  # type: (...) -> None
        self._log_convergence = value
        for mda in self.sub_mda_list:
            mda.log_convergence = value

    def _create_mdo_chain(
        self,
        disciplines,  # type: Sequence[MDODiscipline]
        sub_mda_class="MDAJacobi",  # type: str
        sub_coupling_structures=None,  # type: Optional[Iterable[MDOCouplingStructure]]
        **sub_mda_options  # type: Optional[Union[float,int,bool,str]]
    ):
        """Create an MDO chain from the execution sequence of the disciplines.

        Args:
            sub_mda_class: The name of the class of the sub-MDAs.
            disciplines: The disciplines.
            sub_coupling_structures: The coupling structures to be used by the sub-MDAs.
                If None, they are created from the sub-disciplines.
            **sub_mda_options: The options to be used to initialize the sub-MDAs.
        """
        chained_disciplines = []
        self.sub_mda_list = []

        if sub_coupling_structures is None:
            sub_coupling_structures = repeat(None)

        sub_coupling_structures_iterator = iter(sub_coupling_structures)

        for parallel_tasks in self.coupling_structure.sequence:
            # to parallelize, check if 1 < len(parallel_tasks)
            # for now, parallel tasks are run sequentially
            for coupled_disciplines in parallel_tasks:
                first_disc = coupled_disciplines[0]
                if len(coupled_disciplines) > 1 or (
                    len(coupled_disciplines) == 1
                    and self.coupling_structure.is_self_coupled(first_disc)
                ):
                    # several disciplines coupled

                    # order the MDA disciplines the same way as the
                    # original disciplines
                    sub_mda_disciplines = []
                    for disc in disciplines:
                        if disc in coupled_disciplines:
                            sub_mda_disciplines.append(disc)

                    # create a sub-MDA
                    sub_mda = create_mda(
                        sub_mda_class,
                        sub_mda_disciplines,
                        max_mda_iter=self.max_mda_iter,
                        tolerance=self.tolerance,
                        linear_solver_tolerance=self.linear_solver_tolerance,
                        grammar_type=self.grammar_type,
                        use_lu_fact=self.use_lu_fact,
                        linear_solver=self.linear_solver,
                        linear_solver_options=self.linear_solver_options,
                        coupling_structure=next(sub_coupling_structures_iterator),
                        **sub_mda_options
                    )
                    sub_mda.n_processes = self.n_processes

                    chained_disciplines.append(sub_mda)
                    self.sub_mda_list.append(sub_mda)
                else:
                    # single discipline
                    chained_disciplines.append(first_disc)

        # create the MDO chain that sequentially evaluates the sub-MDAs and the
        # single disciplines
        self.mdo_chain = MDOChain(
            chained_disciplines, name="MDA chain", grammar_type=self.grammar_type
        )

    def _initialize_grammars(self):  # type: (...) -> None
        """Define all inputs and outputs of the chain."""
        if self.mdo_chain is None:  # First call by super class must be ignored.
            return
        self.input_grammar.update_from(self.mdo_chain.input_grammar)
        self.output_grammar.update_from(self.mdo_chain.output_grammar)

    def _check_consistency(self):
        """Check if there is no more than 1 equation per variable.

        For instance if a strong coupling is not also a self coupling.
        """
        if self.mdo_chain is None:  # First call by super class must be ignored.
            return
        super(MDAChain, self)._check_consistency()

    def _run(self):  # type -> None
        if self.warm_start:
            self._couplings_warm_start()
        self.local_data = self.mdo_chain.execute(self.local_data)
        return self.local_data

    def _compute_jacobian(
        self,
        inputs=None,  # type: Optional[Sequence[str]]
        outputs=None,  # type: Optional[Sequence[str]]
    ):  # type: (...) -> None
        if self._chain_linearize:
            self.mdo_chain.add_differentiated_inputs(inputs)
            self.mdo_chain.add_differentiated_outputs(outputs)
            # the Jacobian of the MDA chain is the Jacobian of the MDO chain
            last_cached = self.cache.get_last_cached_inputs()
            self.mdo_chain.linearize(last_cached)
            self.jac = self.mdo_chain.jac
        else:
            super(MDAChain, self)._compute_jacobian(inputs, outputs)

    def add_differentiated_inputs(
        self,
        inputs=None,  # type: Optional[Iterable[str]]
    ):  # type: (...) -> None
        MDA.add_differentiated_inputs(self, inputs)
        if self._chain_linearize:
            self.mdo_chain.add_differentiated_inputs(inputs)

    def add_differentiated_outputs(
        self,
        outputs=None,  # type: Optional[Iterable[str]]
    ):  # type: (...) -> None
        MDA.add_differentiated_outputs(self, outputs=outputs)
        if self._chain_linearize:
            self.mdo_chain.add_differentiated_outputs(outputs)

    @property
    def normed_residual(self):  # type: (...) -> float
        """The normed_residuals, computed from the sub-MDAs residuals."""
        return sum((mda.normed_residual ** 2 for mda in self.sub_mda_list)) ** 0.5

    @normed_residual.setter
    def normed_residual(
        self,
        normed_residual,  # type: float
    ):  # type: (...) ->None
        """Set the normed_residual.

        Has no effect,
        since the normed residuals are defined by sub-MDAs residuals
        (see associated property).

        Here for compatibility with mother class.
        """
        pass

    def get_expected_dataflow(
        self,
    ):  # type: (...) -> List[Tuple[MDODiscipline,MDODiscipline,List[str]]]
        return self.mdo_chain.get_expected_dataflow()

    def get_expected_workflow(self):  # type: (...) ->SerialExecSequence
        exec_s = SerialExecSequence(self)
        workflow = self.mdo_chain.get_expected_workflow()
        exec_s.extend(workflow)
        return exec_s

    def reset_statuses_for_run(self):  # type: (...) -> None
        super(MDAChain, self).reset_statuses_for_run()
        self.mdo_chain.reset_statuses_for_run()

    def plot_residual_history(
        self,
        show=False,  # type: bool
        save=True,  # type: bool
        n_iterations=None,  # type: Optional[int]
        logscale=None,  # type: Optional[Tuple[int,int]]
        filename=None,  # type: Optional[str]
        figsize=(50, 10),  # type: Tuple[int,int]
    ):  # type: (...) -> None
        for sub_mda in self.sub_mda_list:
            if filename is not None:
                s_filename = split(filename)
                filename = join(
                    s_filename[0],
                    "{}_{}".format(sub_mda.__class__.__name__, s_filename[1]),
                )
            sub_mda.plot_residual_history(
                show, save, n_iterations, logscale, filename, figsize
            )



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
# Contributors:
#    INITIAL AUTHORS - API and implementation and/or documentation
#        :author: Francois Gallard
#    OTHER AUTHORS   - MACROSCOPIC CHANGES

"""Base wrapper for all linear solvers."""

import logging
import pickle
from typing import Any, Mapping, Optional
from uuid import uuid4

from numpy import ndarray
from scipy.sparse.linalg import LinearOperator, spilu

from gemseo.algos.algo_lib import AlgoLib
from gemseo.algos.linear_solvers.linear_problem import LinearProblem

LOGGER = logging.getLogger(__name__)


class LinearSolverLib(AlgoLib):
    """Abstract class for libraries of linear solvers.

    Attributes:
        XXX how about backup_path or file_path?
        save_fpath (str): The path to the file to save the problem when
            it is not converged and the attribute save_when_fail is True.
    """

    LHS_MUST_BE_SYMMETRIC = "LHS_symmetric"
    LHS_MUST_BE_POSITIVE_DEFINITE = "LHS_positive_definite"
    LHS_CAN_BE_LINEAR_OPERATOR = "LHS_linear_operator"

    SAVE_WHEN_FAIL = "save_when_fail"

    def __init__(self):  # type: (...) -> None
        super(LinearSolverLib, self).__init__()
        self.save_fpath = None

    def solve(
        self,
        linear_problem,  # type: LinearProblem
        algo_name,  # type: str
        **options  # type: Any
    ):  # type: (...) -> Any
        """Solve the linear system.

        Args:
            linear_problem: The problem to solve.
            algo_name: The name of the algorithm.
            **options: The algorithm options.

        Returns:
            The execution result.
        """
        return self.execute(linear_problem, algo_name=algo_name, **options)

    def _build_ilu_preconditioner(
        self,
        lhs,  # type: ndarray
        dtype=None,  # type: Optional[str]
    ):  # type: (...) -> LinearOperator
        """Construct a preconditioner using an incomplete LU factorization.

        Args:
            lhs: The linear system matrix.
            dtype: The numpy dtype of the resulting linear operator.
                If None, XXX.

        Returns:
            The preconditioner operator.
        """
        ilu = spilu(lhs)
        return LinearOperator(lhs.shape, ilu.solve, dtype=dtype)

    @property
    def solution(self):  # type: (...) -> ndarray
        """The solution of the problem."""
        return self.problem.solution

    @staticmethod
    def is_algorithm_suited(
        algo_dict,  # type: Mapping[str, bool]
        problem,  # type: LinearProblem
    ):  # type: (...) -> bool
        """Check if the algorithm is suited to the problem according to algo_dict.

        Args:
            algo_dict: The algorithm characteristics.
            problem: The problem to be solved.

        Returns:
            Whether the algorithm suits.
        """
        if not problem.is_symmetric and algo_dict.get(
            LinearSolverLib.LHS_MUST_BE_SYMMETRIC, False
        ):
            return False

        if not problem.is_positive_def and algo_dict.get(
            LinearSolverLib.LHS_MUST_BE_POSITIVE_DEFINITE, False
        ):
            return False

        if problem.is_lhs_linear_operator and not algo_dict.get(
            LinearSolverLib.LHS_CAN_BE_LINEAR_OPERATOR, False
        ):
            return False

        return True

    def _pre_run(
        self,
        problem,  # type: LinearProblem
        algo_name,  # type: str
        **options  # type: Any
    ):  # type: (...) -> None
        """Set the solver options and name in the problem attributes.

        Args:
            problem: The problem to be solved.
            algo_name: The name of the algorithm.
            options: The options for the algorithm, see associated JSON file.
        """
        problem.solver_options = options
        problem.solver_name = algo_name

    def _post_run(
        self,
        problem,  # type: LinearProblem
        algo_name,  # type: str
        result,  # type: ndarray
        **options  # type: Any
    ):  # type: (...) -> None # noqa: D107
        """Save the LinearProblem to the disk when required.

        If the save_when_fail option is True, save the LinearProblem to the disk when
        the system failed and print the file name in the warnings.

        Args:
            problem: The problem to be solved.
            algo_name: The name of the algorithm.
            result: The result of the run, i.e. the solution.
            options: The options for the algorithm, see associated JSON file.
        """
        if options.get(self.SAVE_WHEN_FAIL, False):
            if not self.problem.is_converged:
                f_path = "linear_system_{}.pck".format(uuid4())
                pickle.dump(self.problem, open(f_path, "wb"))
                LOGGER.warning(
                    "Linear solver failed, saving problem to file: %s", f_path
                )
                self.save_fpath = f_path



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - API and implementation and/or documentation
#        :author: Francois Gallard
#    OTHER AUTHORS   - MACROSCOPIC CHANGES

"""Wrappers for scipy's linear solvers."""

from __future__ import division, unicode_literals

import logging
from typing import Any, Dict, List, Mapping, Optional, Tuple, Union

from numpy import find_common_type, ndarray
from scipy.sparse import spmatrix
from scipy.sparse.base import issparse
from scipy.sparse.linalg import LinearOperator, bicg, bicgstab, gmres, lgmres, qmr, splu

from gemseo.algos.linear_solvers.linear_solver_lib import LinearSolverLib

LOGGER = logging.getLogger(__name__)


class ScipyLinalgAlgos(LinearSolverLib):
    """Wrapper for scipy linalg sparse linear solvers.

    Attributes:
       save_fpath (str): The path to the file to saved the problem when
           it is not converged and the option save_when_fail
           is active.

       methods_map (Dict[str, Callable]): The mapping between the solver names
           and the solvers methods in scipy.sparse.
    """

    BASE_INFO_MSG = "scipy linear solver algorithm stop info: "
    OPTIONS_MAP = {
        "max_iter": "maxiter",
        "preconditioner": "M",
        "store_outer_av": "store_outer_Av",
    }

    LGMRES_SPEC_OPTS = (
        "inner_m",
        "outer_k",
        "outer_v",
        "store_outer_av",
        "prepend_outer_v",
    )

    __WEBSITE = "https://docs.scipy.org/doc/scipy/reference/generated/{}.html"
    __WEBPAGE = "scipy.sparse.linalg.{}"
    __WEBPAGES = {
        "BICG": __WEBPAGE.format("bicg"),
        "GMRES": __WEBPAGE.format("gmres"),
        "LGMRES": __WEBPAGE.format("lgmres"),
        "QMR": __WEBPAGE.format("qmr"),
        "BICGSTAB": __WEBPAGE.format("bicgstab"),
        "DEFAULT": __WEBPAGE.format("splu"),
    }

    def __init__(self):  # type: (...) -> None
        super(ScipyLinalgAlgos, self).__init__()
        self.methods_map = {
            "LGMRES": lgmres,
            "GMRES": gmres,
            "BICG": bicg,
            "QMR": qmr,
            "BICGSTAB": bicgstab,
            "DEFAULT": self._run_default_solver,
        }
        self.lib_dict = {
            name: self.get_default_properties(name) for name in self.methods_map.keys()
        }
        self.lib_dict["DEFAULT"]["description"] = (
            "This starts by LGMRES, but if it fails, "
            "switches to GMRES, then direct method super LU factorization."
        )

    @classmethod
    def get_default_properties(
        cls, algo_name  # type: str
    ):  # type: (...) -> Dict[str, Union[bool,str]]
        """Return the properties of the algorithm.

        It states if it requires symmetric,
        or positive definite matrices for instance.

        Args:
            algo_name: The algorithm name.

        Returns:
            The properties of the solver.
        """
        return {
            cls.LHS_MUST_BE_POSITIVE_DEFINITE: False,
            cls.LHS_MUST_BE_SYMMETRIC: False,
            cls.LHS_CAN_BE_LINEAR_OPERATOR: True,
            cls.INTERNAL_NAME: algo_name,
            "description": "Linear solver implemented in the SciPy library.",
            "website": cls.__WEBSITE.format(algo_name),
        }

    def _get_options(
        self,
        max_iter=1000,  # type: int
        preconditioner=None,  # type: Optional[Union[ndarray, LinearOperator]]
        tol=1e-12,  # type: float
        atol=None,  # type: Optional[float]
        x0=None,  # type: Optional[ndarray]
        use_ilu_precond=True,  # type: bool
        inner_m=30,  # type: int
        outer_k=3,  # type: int
        outer_v=None,  # type: Optional[List[Tuple]]
        store_outer_av=True,  # type: bool
        prepend_outer_v=False,  # type: bool
        save_when_fail=False,  # type: bool
        store_residuals=False,  # type: bool
    ):  # type: (...) -> Dict[str, Any]
        """Check the options and sets the default values.

        See https://docs.scipy.org/doc/scipy/reference/sparse.linalg.html

        Args:
            max_iter: The maximum number of iterations.
            preconditioner: The preconditionner, approximation of RHS^-1.
                If None, no preconditioner is used.
            tol: The relative tolerance for convergence,
                norm(RHS.dot(sol)) <= max(tol*norm(LHS), atol).
            atol: The absolute tolerance for convergence,
                norm(RHS.dot(sol)) <= max(tol*norm(LHS), atol).
            x0: The initial guess for the solution.
                M{sparse matrix, dense matrix, LinearOperator}.
                If None, solvers usually start from the null vector.
            inner_m int: The number of inner GMRES iterations per outer iteration.
            outer_k: The number of vectors to carry between inner GMRES iterations.
            outer_v:  The data used to augment the Krylov subspace.
            store_outer_A: Whether LGMRES should store also A*v in
                addition to the vectors v in outer_v.
            prepend_outer_v: Whether to put outer_v
                augmentation vectors before the Krylov iterates.
            use_ilu_precond: Whether to use superLU to
                compute an incomplete LU factorization as preconditioner.
            save_when_fail: Whether to save the linear system to the disk
                when the solver failed to converge.
            store_residuals: Whether to store the residuals convergence history.

        Returns:
            The options.

        Raises:
            ValueError: When the LHR and RHS shapes are inconsistent, or
                when the preconditioner options are inconsistent.
        """
        if preconditioner is not None and (
            preconditioner.shape != self.problem.lhs.shape
        ):
            msg = "Inconsistent Preconditioner shape: %s != %s"
            raise ValueError(msg.format(preconditioner.shape, self.problem.lhs.shape))

        if x0 is not None and len(x0) != len(self.problem.rhs):
            msg = "Inconsistent initial guess shape: %s != %s"
            raise ValueError(msg.format(x0.shape, self.problem.rhs.shape))

        if use_ilu_precond and preconditioner is not None:
            raise ValueError(
                "Use either 'use_ilu_precond' or provide 'preconditioner',"
                " but not both."
            )

        return self._process_options(
            max_iter=max_iter,
            preconditioner=preconditioner,
            tol=tol,
            atol=atol,
            x0=x0,
            inner_m=inner_m,
            outer_k=outer_k,
            outer_v=outer_v,
            store_outer_av=store_outer_av,
            prepend_outer_v=prepend_outer_v,
            use_ilu_precond=use_ilu_precond,
            save_when_fail=save_when_fail,
            store_residuals=store_residuals,
        )

    def _run(
        self, **options  # type: Union[None, bool, int, float, ndarray]
    ):  # type: (...) -> ndarray
        """Run the algorithm.

        Args:
            **options: The options for the algorithm.

        Returns:
            The solution of the problem.
        """
        if issparse(self.problem.rhs):
            self.problem.rhs = self.problem.rhs.toarray()
        rhs = self.problem.rhs
        lhs = self.problem.lhs

        opts_solver = options.copy()
        c_dtype = None

        if rhs.dtype != lhs.dtype and not isinstance(lhs, LinearOperator):
            c_dtype = find_common_type([rhs.dtype, lhs.dtype], [])
            if lhs.dtype != c_dtype:
                lhs = lhs.astype(c_dtype)
            if rhs.dtype != c_dtype:
                rhs = rhs.astype(c_dtype)

        if opts_solver["use_ilu_precond"] and not isinstance(lhs, LinearOperator):
            opts_solver["M"] = self._build_ilu_preconditioner(lhs, c_dtype)

        del opts_solver["use_ilu_precond"]
        del opts_solver["store_residuals"]

        method = self.methods_map[self.algo_name]

        if options.get("store_residuals", False):
            opts_solver["callback"] = self.__store_residuals

        opts_solver.pop("save_when_fail")
        self.problem.solution, info = method(lhs, rhs, **opts_solver)
        self._check_solver_info(info, options)

        return self.problem.solution

    def __store_residuals(
        self, current_x  # type: ndarray
    ):  # type: (...) -> ndarray
        """Store the current iteration residuals.

        Args:
            current_x: The current solution.
        """
        self.problem.solution = current_x
        self.problem.compute_residuals(True, True)

    def _check_solver_info(
        self,
        info,  # type: int
        options,  # type: Mapping[str, Any]
    ):  # type: (...) -> bool
        """Check the info returned by the solver.

        Args:
            info: The info value, negative, 0 or positive depending
                on status.
            options: The options passed to the solver.

        Returns:
            Whether the solver converged.

        Raises:
            RuntimeError: If the inputs are illegal for the solver.
        """
        self.problem.is_converged = info == 0

        if info > 0:
            if self.problem.solution is not None:
                LOGGER.warning(
                    "%s, residual = %s",
                    self.BASE_INFO_MSG,
                    self.problem.compute_residuals(True),
                )
                LOGGER.warning("info = %s", info)
            else:
                LOGGER.warning("Failed to converge")
            return False

        # check the dimensions
        if info < 0:
            raise RuntimeError(
                self.BASE_INFO_MSG + "illegal input or breakdown" ", options = %s",
                options,
            )

        return True

    def _run_default_solver(
        self,
        lhs,  # type: Union[ndarray, spmatrix, LinearOperator]
        rhs,  # type: ndarray
        **options  # type: Any
    ):  # type: (...) -> Tuple[ndarray, int]
        """Run the default solver.

        This starts by LGMRES, but if it fails, switches to GMRES,
        then direct method super LU factorization.

        Args:
            lhs: The left hand side of the equation (matrix).
            rhs: The right hand side of the equation.
            **options: The user options.

        Returns:
            The last solution found and the info.
        """
        # Try LGMRES first
        best_sol, info = lgmres(A=lhs, b=rhs, **options)
        min_res = self.problem.compute_residuals(True, current_x=best_sol)

        if self._check_solver_info(info, options):
            # If converged, stop
            return best_sol, info
        else:
            # Otherwise try GMRES
            min_res = self.problem.compute_residuals(True, current_x=best_sol)
            for k in self.LGMRES_SPEC_OPTS:  # Adapt options
                if k in options:
                    del options[k]

            if min_res < 1.0:
                options["x0"] = best_sol

            sol, info = gmres(A=lhs, b=rhs, **options)
            res = self.problem.compute_residuals(True, current_x=sol)

            if res < min_res:
                best_sol = sol
                options["x0"] = best_sol

            if self._check_solver_info(info, options):  # pragma: no cover
                return best_sol, info

        # In this case previous runs failed, trying direct method
        # based on super LU
        a_fact = splu(lhs)
        sol = a_fact.solve(rhs)
        res = self.problem.compute_residuals(True, current_x=best_sol)

        if res < options["tol"]:  # pragma: no cover
            best_sol = sol
            info = 0
            self.problem.is_converged = True
        else:
            info = 1

        return best_sol, info



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                         documentation
#        :author: Francois Gallard
#    OTHER AUTHORS   - MACROSCOPIC CHANGES

"""Rules and checks for disciplines inputs/outputs validation."""

from __future__ import division, unicode_literals

import logging
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Iterable,
    List,
    Mapping,
    Optional,
    Sequence,
    Union,
)

if TYPE_CHECKING:
    from gemseo.core.grammars.simple_grammar import SimpleGrammar

import six
from custom_inherit import DocInheritMeta
from numpy import ndarray, zeros

from gemseo.utils.py23_compat import Path

LOGGER = logging.getLogger(__name__)


@six.add_metaclass(
    DocInheritMeta(
        abstract_base_class=True,
        style="google_with_merge",
        include_special_methods=True,
    )
)
class AbstractGrammar(object):
    """The abstraction of a grammar.

    A grammar is a set of elements characterised by their names and types.
    One can use it to check if elements values are consistent with it.

    It is mainly used to store the names and types
    of the inputs and outputs of an :class:`.MDODiscipline`.

    Grammars supports __contains__ special method, so that one can test
    if an element is in the grammar with the statement `"x" in grammar`.

    Attributes:
        name (str): The name of the grammar.
    """

    def __init__(
        self,
        name,  # type: str
        **kwargs  # type: Union[str,Path]
    ):  # type: (...) -> None
        """
        Args:
            name: The name to be given to the grammar.
            **kwargs: The options of the grammar.
        """
        self.name = name

    def __str__(self):  # type: (...) -> str
        return "grammar name: {}".format(self.name)

    def __contains__(
        self,
        item,  # type: str
    ):  # type: (...) -> bool
        return self.is_data_name_existing(item)

    def load_data(
        self,
        data,  # type: Mapping[str,Any]
        raise_exception=True,  # type: bool
    ):  # type: (...) -> Mapping[str,Any]
        """Load elements values and check their consistency with the grammar.

        Args:
            data: The elements values to be checked.
            raise_exception: Whether to raise an exception
                when the elements values are not consistent with the grammar.

        Returns:
            The elements values after successful consistency checking.
        """
        raise NotImplementedError()

    def get_data_names(self):  # type: (...) -> List[str]
        """Return the names of the elements.

        Returns:
            The names of the elements sorted alphabetically.
        """
        raise NotImplementedError()

    def update_from(
        self,
        input_grammar,  # type: AbstractGrammar
    ):  # type: (...) -> None
        """Update the grammar with a second one.

        Add the new elements and update the existing ones.

        Args:
            input_grammar: The grammar to take the elements from.
        """
        raise NotImplementedError()

    def is_type_array(
        self, data_name  # type: str
    ):  # type: (...) -> bool
        """Check if an element is an array.

        Args:
            data_name: The name of the element.

        Returns:
            Whether the element is an array.

        Raises:
            ValueError: If the name does not correspond to an element name.
        """
        raise NotImplementedError()

    def update_from_if_not_in(
        self,
        input_grammar,  # type: AbstractGrammar
        exclude_grammar,  # type: AbstractGrammar
    ):  # type: (...) -> None
        """Add the elements from a second grammar that are not present in a third one.

        Args:
            input_grammar: The grammar to take the elements from.
            exclude_grammar: The grammar containing the elements not to be taken.
        """
        raise NotImplementedError()

    def is_data_name_existing(
        self,
        data_name,  # type: str
    ):  # type: (...) -> bool
        """Check if the name of an element is present in the grammar.

        Args:
            data_name: The name of the element.

        Returns:
            Whether the name of the element is present in the grammar.
        """
        raise NotImplementedError()

    def is_all_data_names_existing(
        self,
        data_names,  # type: Iterable[str]
    ):  # type: (...) -> bool
        """Check if the names of the elements are present in the grammar.

        Args:
            data_names: The names of the elements.

        Returns:
            Whether all the elements names are in the grammar.
        """
        raise NotImplementedError()

    def clear(self):  # type: (...) -> None
        """Clear the grammar."""
        raise NotImplementedError()

    def to_simple_grammar(self):  # type: (...) -> SimpleGrammar
        """Convert to the base :class:`.SimpleGrammar` type.

        Returns:
            A :class:`.SimpleGrammar` version of the current grammar.
        """
        raise NotImplementedError()

    def initialize_from_data_names(
        self,
        data_names,  # type: Iterable[str]
    ):  # type: (...) -> None
        """Initialize the grammar from the names of the elements and float type.

        Args:
            data_names: The names of the elements.
        """
        element_value = zeros(1)
        elements_values = {element_name: element_value for element_name in data_names}
        self.initialize_from_base_dict(elements_values)

    def initialize_from_base_dict(
        self,
        typical_data_dict,  # type: Dict[str,ndarray]
    ):  # type: (...) -> None
        """Initialize the grammar with types and names from typical elements values.

        Args:
            typical_data_dict: Typical elements values
                indexed by the elements names.
        """
        raise NotImplementedError()

    @staticmethod
    def _get_update_error_msg(
        grammar1,  # type: AbstractGrammar
        grammar2,  # type: AbstractGrammar
        grammar3=None,  # type: Optional[AbstractGrammar]
    ):  # type: (...) -> str
        """Create a message for grammar update error.

        Args:
            grammar1: The grammar to be updated.
            grammar2: A grammar to update the first one.
            grammar3: Another grammar to update the first one.
                If None,
                consider that only the ``grammar2`` is used to update ``grammar1``.

        Returns:
            The error message based on the passed grammars.
        """
        msg = "Cannot update grammar {} of type {} with {} of type {}".format(
            grammar1.name,
            grammar1.__class__.__name__,
            grammar2.name,
            grammar2.__class__.__name__,
        )
        if grammar3 is not None:
            msg += " and {} of type {}".format(
                grammar1.name,
                grammar1.__class__.__name__,
            )
        return "{}.".format(msg)

    def restrict_to(
        self,
        data_names,  # type: Sequence[str]
    ):  # type: (...) -> None
        """Restrict the grammar to the given names.

        Args:
            data_names: The names of the elements to restrict the grammar to.
        """
        raise NotImplementedError

    def remove_item(
        self,
        item_name,  # type: str
    ):  # type: (...) -> None
        """Remove an element.

        Args:
            item_name: The name of the element to be removed.

        Raises:
            KeyError: When the element is not in the grammar.
        """
        raise NotImplementedError

    def get_type_from_python_type(
        self, python_type  # type: type
    ):  # type: (...) -> type
        """Return the grammar type that corresponds to a given Python type.

        Args:
            python_type: The Python type.

        Return:
            The equivalent grammar type.
        """
        raise NotImplementedError

    def update_elements(
        self,
        python_typing=False,  # type: bool
        **elements  # type: Mapping[str,type]
    ):  # type: (...) -> None
        """Add or update elements from their names and types.

        Args:
            python_typing: If True, handle automatically the conversion from
                Python type to grammar type.
            **elements: The names to types bindings of the elements to add or update.

        Examples:
            >>> grammar.update_elements(a=str, b=int)
            >>> grammar.update_elements(a=str, b=int, python_typing=True)
            >>> grammar.update_elements(**names_to_types)
        """
        raise NotImplementedError

    def update_required_elements(
        self, **elements  # type: Mapping[str, bool]
    ):  # type: (...) -> None
        """Add or update the required elements in the grammar.

        Args:
            **elements: The names of the elements bound to whether they shall be required.

        Raises:
            KeyError: If a given element name is not in the grammar.
            TypeError: If a given element name is not associated to a boolean value.
        """
        raise NotImplementedError

    def is_required(
        self, element_name  # type: str
    ):  # type: (...) -> bool
        """Check if an element is required in the grammar.

        Args:
            element_name: The data name to check.

        Returns:
            Whether the element name is required.

        Raises:
            ValueError: If the given element is not in the grammar.
        """
        raise NotImplementedError
